#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re

from loguru import logger
from tqdm import tqdm


def generate_brat_annotation_conf(ner_labels, ner_connections):
    annotation_conf_text = "[entities]\n"
    for x in ner_labels:
        annotation_conf_text += f"{x}\n"
    annotation_conf_text += "\n"

    annotation_conf_text += "[relations]\n"
    #  for c, from_entity, to_entity in ner_connections:
    #      annotation_conf_text += "f{c} {from_entity},{to_entity}\n"
    annotation_conf_text += "\n"

    annotation_conf_text += "[events]\n"

    annotation_conf_text += "\n"

    annotation_conf_text += "[attributes]\n"

    annotation_conf_text += "\n"

    return annotation_conf_text


def generate_brat_visual_conf(ner_labels):
    visual_conf_text = "[labels]\n"
    for x in ner_labels:
        visual_conf_text += f"{x} | {x}\n"
    visual_conf_text += "\n"

    brat_bgcolors = [
        '#007bff', '#17a2b8', '#28a745', '#fd7e14', '#e83e8c', '#dc3545',
        '#20c997', '#ffc107', '#007bff'
    ]
    #  [
    #      '#ffccaa', '#8fb2ff', '#7fe2ff', '#6fffdf', '#aaaaee', '#ccccee',
    #      'darkgray', '#007000', '#7fa2ff', '#cf9fff', '#9fc2ff', '#e0ff00',
    #      '#18c59a', '#b4c8ff', '#ffff00'
    #  ]

    brat_fdcolors = []

    visual_conf_text += "[drawing]\n"
    visual_conf_text += "SPAN_DEFAULT\tfgColor:black, bgColor:lightgreen, borderColor:darken\n"
    visual_conf_text += "ARC_DEFAULT\tcolor:black, arrowHead:triangle-5\n"
    visual_conf_text += "\n"
    visual_conf_text += "Alias\tdashArray:3-3, arrowHead:none\n"
    visual_conf_text += "Equiv\tdashArray:3-3, arrowHead:none\n"
    visual_conf_text += "\n"
    for i, x in enumerate(ner_labels):
        visual_conf_text += f"{x} bgColor:{brat_bgcolors[i % len(brat_bgcolors)]}\n"
    visual_conf_text += "\n"

    visual_conf_text += "# Attributes\n"
    visual_conf_text += "ATTRIBUTE_DEFAULT\tglyph:*\n"
    visual_conf_text += "Negation\tbox:crossed, glyph:<NONE>, dashArray:<NONE>\n"
    visual_conf_text += "Confidence\tglyph:↑|↔|↓\n"
    visual_conf_text += "Speculation\tdashArray:3-3, glyph:<NONE>\n"

    visual_conf_text += "[options]\n"
    visual_conf_text += "Arcs bundle:none\n"
    visual_conf_text += "Text direction:ltr\n"
    visual_conf_text += "\n"

    return visual_conf_text


def generate_brat_kb_shortcuts_conf(ner_labels):
    kb_shortcuts_conf_text = ""
    for i, x in enumerate(ner_labels):
        kb_shortcuts_conf_text += f"{i}\t{x}\n"
        if i >= 9:
            break
    return kb_shortcuts_conf_text


def save_brat_file(tagged_text_list,
                   brat_data_dir,
                   ner_labels,
                   ner_connections,
                   start_page=0,
                   max_pages=10):

    if len(tagged_text_list) == 0:
        return

    assert max_pages >= 1
    if max_pages == 1:
        brat_data_file = tagged_text_list[start_page].guid
        p0 = brat_data_file.rfind('.')
        if p0 > 0:
            brat_data_file = brat_data_file[:p0]
        brat_data_file = os.path.join(brat_data_dir, brat_data_file)
    else:
        brat_data_file = os.path.join(brat_data_dir,
                                      f"{max_pages}_{start_page}")

    brat_colorset = [
        '#007bff', '#17a2b8', '#28a745', '#fd7e14', '#e83e8c', '#dc3545',
        '#20c997', '#ffc107', '#007bff'
    ]
    label2id = {x: i for i, x in enumerate(ner_labels)}

    #  for _id, _text in enumerate(ner_connections):
    #      connection_categories.append({'id': _id, 'text': _text})

    brat_content = ""
    brat_ann_content = ""
    num_pages = 0
    page_offset = 0

    eid = 0
    for tagged_text in tqdm(tagged_text_list, desc="brat files"):
        if num_pages < start_page:
            num_pages += 1
            continue
        guid = tagged_text.guid
        text = tagged_text.text

        if max_pages <= 1:
            page_head = ""
            page_tail = ""
        else:
            page_head = f"\n-------------------- {guid} Begin --------------------\n\n"
            page_tail = f"\n-------------------- {guid} End --------------------\n\n"
        brat_content += page_head + f"{text}" + page_tail

        tags = sorted(tagged_text.tags, key=lambda x: x.start)
        for entity_tag in tags:
            category = entity_tag.category
            mention = entity_tag.mention
            start_char = entity_tag.start
            end_char = start_char + len(mention)

            start_char = page_offset + len(page_head) + start_char
            end_char = page_offset + len(page_head) + end_char

            brat_ann_line = f"T{eid+1}\t{category} {start_char} {end_char}\t{mention}\n"
            brat_ann_content += brat_ann_line

            eid += 1

        num_pages += 1
        if num_pages - start_page >= max_pages:
            break

        page_offset = len(brat_content)

    logger.info(f"Saved {brat_data_file}")

    with open(f"{brat_data_file}.txt", 'w') as wt:
        wt.write(brat_content)

    with open(f"{brat_data_file}.ann", 'w') as wt:
        wt.write(brat_ann_content)


def export_brat_files(tagged_text_list,
                      ner_labels,
                      ner_connections,
                      brat_data_dir,
                      max_pages=10):
    import os, shutil
    if os.path.exists(brat_data_dir):
        shutil.rmtree(brat_data_dir)
    os.makedirs(brat_data_dir)

    annotation_conf_text = generate_brat_annotation_conf(
        ner_labels, ner_connections)
    visual_conf_text = generate_brat_visual_conf(ner_labels)
    kb_shortcuts_conf_text = generate_brat_kb_shortcuts_conf(ner_labels)

    with open(os.path.join(brat_data_dir, "annotation.conf"), 'w') as wt:
        wt.write(annotation_conf_text)

    with open(os.path.join(brat_data_dir, "visual.conf"), 'w') as wt:
        wt.write(visual_conf_text)

    with open(os.path.join(brat_data_dir, "kb_shortcuts.conf"), 'w') as wt:
        wt.write(kb_shortcuts_conf_text)

    num_pages = len(tagged_text_list)
    for start_page in range(0, num_pages, max_pages):
        save_brat_file(tagged_text_list,
                       brat_data_dir,
                       ner_labels,
                       ner_connections,
                       start_page=start_page,
                       max_pages=max_pages)


def get_brat_schemas(brat_data_dir):
    ner_labels = []
    ner_connections = []

    import glob
    txt_files = glob.glob(os.path.join(brat_data_dir, "*.txt"))
    txt_files = sorted(txt_files)
    ann_files = [x[:-3] + "ann" for x in txt_files]

    for txt_file, ann_file in tqdm(zip(txt_files, ann_files),
                                   desc="Brat schemas"):
        #  logger.info(f"{txt_file}")

        content = open(txt_file, 'r').read()
        ann_content = open(ann_file, 'r').read()

        tags = []
        for ann_line in ann_content.split('\n'):
            ann_line = ann_line.strip()
            if len(ann_line) == 0:
                continue
            #  logger.info(f"ann_line: {ann_line}")
            toks = ann_line.split('\t')
            #  logger.info(f"toks: {toks}")
            tid, label, mention = ann_line.split('\t')
            c, s, e = label.split(' ')
            ner_labels.append(c)

    ner_labels = list(set(ner_labels))
    ner_connections = list(set(ner_connections))

    return ner_labels, ner_connections


def brat_data_generator(brat_data_dir):

    import glob
    txt_files = glob.glob(os.path.join(brat_data_dir, "*.txt"))
    txt_files = sorted(txt_files)
    ann_files = [x[:-3] + "ann" for x in txt_files]

    for txt_file, ann_file in tqdm(zip(txt_files, ann_files),
                                   desc="Load from brat"):
        #  logger.info(f"{txt_file}")
        guid = os.path.basename(txt_file)

        content = open(txt_file, 'r').read()
        ann_content = open(ann_file, 'r').read()

        tags = []
        for ann_line in ann_content.split('\n'):
            ann_line = ann_line.strip()
            if len(ann_line) == 0:
                continue
            #  logger.info(f"ann_line: {ann_line}")
            toks = ann_line.split('\t')
            #  logger.info(f"toks: {toks}")
            tid, label, mention = ann_line.split('\t')
            c, s, e = label.split(' ')
            s = int(s)
            e = int(e)
            tags.append([c, s, mention])

        re_b = '(\\n[-]+ (.*?) Begin [-]+\\n\\n)'
        re_e = '(\\n[-]+ (.*?) End [-]+\\n\\n)'

        b_list = []
        for x in re.finditer(re_b, content):
            b_list.append((x.start(), x.end(), x.groups()[1]))
        e_list = []
        for x in re.finditer(re_e, content):
            e_list.append((x.start(), x.end()))

        pages = [(x_b[2], x_b[0], x_b[1], x_e[0], x_e[1])
                 for x_b, x_e in zip(b_list, e_list)]

        #  logger.warning(f"pages: {pages}")
        if pages:
            for i, page in enumerate(pages):
                guid, head_x0, head_x1, tail_x0, tail_x1 = page

                text = content[head_x1:tail_x0]

                json_tags = []
                for c, s, m in tags:
                    e = s + len(m) - 1
                    assert e >= s
                    if s >= head_x1 and e < tail_x0:

                        json_tags.append({
                            'category': c,
                            'start': s - head_x1,
                            'mention': m,
                        })
                json_tags = sorted(json_tags, key=lambda x: x['start'])
                yield guid, text, None, json_tags
        else:
            text = content
            json_tags = []
            for c, s, m in tags:
                e = s + len(m) - 1
                assert e >= s

                json_tags.append({
                    'category': c,
                    'start': s,
                    'mention': m,
                })
            json_tags = sorted(json_tags, key=lambda x: x['start'])
            yield guid, text, None, json_tags



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import json
from pathlib import Path

import numpy as np
import torch
from loguru import logger
from torch.utils.data import (DataLoader, RandomSampler, SequentialSampler,
                              TensorDataset)
from tqdm import tqdm

from ...utils import seg_generator
from ...utils.multiprocesses import (barrier_leader_process,
                                     barrier_member_processes,
                                     is_multi_processes)
from ..ner_utils import InputExample
from ..trainer import common_batch_encode, common_to_tensors


class InputFeature(object):
    """A single set of features of data."""
    def __init__(self, input_ids, attention_mask, token_type_ids, input_len,
                 token_offsets, text, labels, subjects):
        self.input_ids = input_ids
        self.attention_mask = attention_mask
        self.token_type_ids = token_type_ids
        self.input_len = input_len
        self.token_offsets = token_offsets
        self.text = text
        self.labels = labels
        self.subjects = subjects

    #  def __repr__(self):
    #      return str(self.to_json_string())
    #
    #  def __eq__(self, other):
    #      return self.input_ids == other.input_ids and \
    #          self.attention_mask == other.attention_mask and \
    #          self.token_type_ids == other.token_type_ids and \
    #          self.labels == other.labels and \
    #          self.subjects == other.subjects and \
    #          self.input_len == other.input_len
    #
    #  def to_dict(self):
    #      """Serializes this instance to a Python dictionary."""
    #      output = copy.deepcopy(self.__dict__)
    #      return output
    #
    #  def to_json_string(self):
    #      """Serializes this instance to a JSON string."""
    #      return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


#  def batch_to_input_data(batch):
#      all_input_ids = torch.stack([e.input_ids for e in batch])
#      all_attention_mask = torch.stack([e.attention_mask for e in batch])
#      all_token_type_ids = torch.stack([e.token_type_ids for e in batch])
#      all_labels = torch.stack([e.labels for e in batch])
#      all_input_lens = torch.stack([e.input_len for e in batch])
#      all_token_offsets = torch.stack([e.token_offsets for e in batch])
#      all_subjects = [e.subjects for e in batch]
#
#      return (all_input_ids, all_attention_mask, all_token_type_ids, all_labels,
#              all_input_lens, all_token_offsets, all_subjects)


def encode_examples(examples, label2id, tokenizer, max_seq_length):
    logger.info(f"Encoding examples...")

    num_labels = len(label2id)
    texts = [e.text_a[:max_seq_length - 2] for e in examples]

    all_tokens, all_token2char, all_char2token, all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets = common_batch_encode(
        texts, label2id, tokenizer, max_seq_length)

    def encode_subjects(subjects, num_tokens, char2token, input_len):

        subjects_id = []
        labels = np.zeros((num_tokens, num_labels * 2), dtype=int).tolist()

        if subjects:
            for subject in subjects:
                #  logger.debug(f"subject: {subject}")
                label = subject[0]
                start = subject[1]
                end = subject[2]

                if start < 0 or start >= len(char2token):
                    continue
                if end < 0 or end >= len(char2token):
                    continue
                start = char2token[start]
                end = char2token[end]

                if start < input_len - 1 and end < input_len:
                    labels[start][(label2id[label] - 1) * 2] = 1
                    labels[end][(label2id[label] - 1) * 2 + 1] = 1

                    subjects_id.append((label2id[label], start, end))

        return subjects_id, labels

    all_subjects_ids = []
    all_labels = []
    all_subjects = [e.labels for e in examples]
    for input_ids, attention_mask, token_type_ids, subjects, char2token, input_len in tqdm(
            zip(all_input_ids, all_attention_mask, all_token_type_ids,
                all_subjects, all_char2token, all_input_lens),
            desc="encode_subjects"):
        num_tokens = len(input_ids)
        subjects_id, labels = encode_subjects(subjects, num_tokens, char2token,
                                              input_len)

        assert len(input_ids) == max_seq_length
        assert len(attention_mask) == max_seq_length
        assert len(token_type_ids) == max_seq_length
        assert len(labels) == max_seq_length

        all_subjects_ids.append(subjects_id)
        all_labels.append(labels)

    logger.debug(f"all_input_ids.shape: {all_input_ids.shape}")
    logger.debug(f"all_attention_mask.shape: {all_attention_mask.shape}")
    logger.debug(f"all_token_type_ids.shape: {all_token_type_ids.shape}")
    logger.debug(f"all_input_lens.shape: {all_input_lens.shape}")
    logger.debug(f"all_labels.shape: {len(all_labels)}")
    logger.debug(f"all_subjects_ids.shape: {len(all_subjects_ids)}")
    assert all_input_ids.shape[1] == max_seq_length
    assert all_attention_mask.shape[1] == max_seq_length
    assert all_token_type_ids.shape[1] == max_seq_length
    #  logger.debug(f"all_input_ids.shape: {np.array(all_input_ids).shape}")
    #  logger.debug(
    #      f"all_attention_mask.shape: {np.array(all_attention_mask).shape}")
    #  logger.debug(
    #      f"all_token_type_ids.shape: {np.array(all_token_type_ids).shape}")
    #  logger.debug(f"all_labels.shape: {np.array(all_labels).shape}")
    #  logger.debug(f"all_subjects_ids.shape: {np.array(all_subjects_ids).shape}")
    #  logger.debug(f"all_input_lens.shape: {np.array(all_input_lens).shape}")
    #  assert np.array(all_input_ids).shape[1] == max_seq_length
    #  assert np.array(all_attention_mask).shape[1] == max_seq_length
    #  assert np.array(all_token_type_ids).shape[1] == max_seq_length

    all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets = common_to_tensors(
        all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens,
        all_token_offsets)
    #  all_input_ids = torch.from_numpy(np.array(all_input_ids, dtype=np.int64))
    #  all_attention_mask = torch.from_numpy(
    #      np.array(all_attention_mask, dtype=np.int64))
    #  all_token_type_ids = torch.from_numpy(
    #      np.array(all_token_type_ids, dtype=np.int64))
    #  all_input_lens = torch.from_numpy(np.array(all_input_lens, dtype=np.int64))
    #  all_token_offsets = torch.from_numpy(
    #      np.array(all_token_offsets, dtype=np.int64))

    all_labels = torch.from_numpy(np.array(all_labels, dtype=np.int64))
    all_subjects_ids = all_subjects_ids

    all_features = [
        InputFeature(input_ids=input_ids,
                     attention_mask=attention_mask,
                     token_type_ids=token_type_ids,
                     input_len=input_len,
                     token_offsets=token_offsets,
                     text=text,
                     labels=labels,
                     subjects=subjects_ids)
        for input_ids, attention_mask, token_type_ids, input_len,
        token_offsets, text, labels, subjects_ids in tqdm(
            zip(all_input_ids, all_attention_mask, all_token_type_ids,
                all_input_lens, all_token_offsets, texts, all_labels,
                all_subjects_ids),
            desc="build all_features")
    ]

    return all_features



# -*- coding: utf-8 -*-
# File: bsds500.py


import glob
import numpy as np
import os

from ...utils.fs import download, get_dataset_path
from ..base import RNGDataFlow

__all__ = ['BSDS500']


DATA_URL = "http://www.eecs.berkeley.edu/Research/Projects/CS/vision/grouping/BSR/BSR_bsds500.tgz"
DATA_SIZE = 70763455
IMG_W, IMG_H = 481, 321


class BSDS500(RNGDataFlow):
    """
    `Berkeley Segmentation Data Set and Benchmarks 500 dataset
    <http://www.eecs.berkeley.edu/Research/Projects/CS/vision/grouping/resources.html#bsds500>`_.

    Produce ``(image, label)`` pair, where ``image`` has shape (321, 481, 3(BGR)) and
    ranges in [0,255].
    ``Label`` is a floating point image of shape (321, 481) in range [0, 1].
    The value of each pixel is ``number of times it is annotated as edge / total number of annotators for this image``.
    """

    def __init__(self, name, data_dir=None, shuffle=True):
        """
        Args:
            name (str): 'train', 'test', 'val'
            data_dir (str): a directory containing the original 'BSR' directory.
        """
        # check and download data
        if data_dir is None:
            data_dir = get_dataset_path('bsds500_data')
        if not os.path.isdir(os.path.join(data_dir, 'BSR')):
            download(DATA_URL, data_dir, expect_size=DATA_SIZE)
            filename = DATA_URL.split('/')[-1]
            filepath = os.path.join(data_dir, filename)
            import tarfile
            tarfile.open(filepath, 'r:gz').extractall(data_dir)
        self.data_root = os.path.join(data_dir, 'BSR', 'BSDS500', 'data')
        assert os.path.isdir(self.data_root)

        self.shuffle = shuffle
        assert name in ['train', 'test', 'val']
        self._load(name)

    def _load(self, name):
        image_glob = os.path.join(self.data_root, 'images', name, '*.jpg')
        image_files = glob.glob(image_glob)
        gt_dir = os.path.join(self.data_root, 'groundTruth', name)
        self.data = np.zeros((len(image_files), IMG_H, IMG_W, 3), dtype='uint8')
        self.label = np.zeros((len(image_files), IMG_H, IMG_W), dtype='float32')

        for idx, f in enumerate(image_files):
            im = cv2.imread(f, cv2.IMREAD_COLOR)
            assert im is not None
            if im.shape[0] > im.shape[1]:
                im = np.transpose(im, (1, 0, 2))
            assert im.shape[:2] == (IMG_H, IMG_W), "{} != {}".format(im.shape[:2], (IMG_H, IMG_W))

            imgid = os.path.basename(f).split('.')[0]
            gt_file = os.path.join(gt_dir, imgid)
            gt = loadmat(gt_file)['groundTruth'][0]
            n_annot = gt.shape[0]
            gt = sum(gt[k]['Boundaries'][0][0] for k in range(n_annot))
            gt = gt.astype('float32')
            gt *= 1.0 / n_annot
            if gt.shape[0] > gt.shape[1]:
                gt = gt.transpose()
            assert gt.shape == (IMG_H, IMG_W)

            self.data[idx] = im
            self.label[idx] = gt

    def __len__(self):
        return self.data.shape[0]

    def __iter__(self):
        idxs = np.arange(self.data.shape[0])
        if self.shuffle:
            self.rng.shuffle(idxs)
        for k in idxs:
            yield [self.data[k], self.label[k]]


try:
    from scipy.io import loadmat
    import cv2
except ImportError:
    from ...utils.develop import create_dummy_class
    BSDS500 = create_dummy_class('BSDS500', ['scipy.io', 'cv2'])  # noqa

if __name__ == '__main__':
    a = BSDS500('val')
    a.reset_state()
    for k in a:
        cv2.imshow("haha", k[1].astype('uint8') * 255)
        cv2.waitKey(1000)



# -*- coding: utf-8 -*-
# File: base.py

import os
import inspect
import pprint
from collections import namedtuple
import weakref

from ...utils.argtools import log_once
from ...utils.utils import get_rng
from ...utils.develop import deprecated
from ..image import check_dtype

# Cannot import here if we want to keep backward compatibility.
# Because this causes circular dependency
# from .transform import TransformList, PhotometricTransform, TransformFactory

__all__ = ['Augmentor', 'ImageAugmentor', 'AugmentorList', 'PhotometricAugmentor']


def _reset_augmentor_after_fork(aug_ref):
    aug = aug_ref()
    if aug:
        aug.reset_state()


def _default_repr(self):
    """
    Produce something like:
    "imgaug.MyAugmentor(field1={self.field1}, field2={self.field2})"

    It assumes that the instance `self` contains attributes that match its constructor.
    """
    classname = type(self).__name__
    argspec = inspect.getfullargspec(self.__init__)
    assert argspec.varargs is None, "The default __repr__ in {} doesn't work for varargs!".format(classname)
    assert argspec.varkw is None, "The default __repr__ in {} doesn't work for kwargs!".format(classname)
    defaults = {}

    fields = argspec.args[1:]
    defaults_pos = argspec.defaults
    if defaults_pos is not None:
        for f, d in zip(fields[::-1], defaults_pos[::-1]):
            defaults[f] = d

    for k in argspec.kwonlyargs:
        fields.append(k)
        if k in argspec.kwonlydefaults:
            defaults[k] = argspec.kwonlydefaults[k]

    argstr = []
    for f in fields:
        assert hasattr(self, f), \
            "Attribute {} in {} not found! Default __repr__ only works if " \
            "the instance has attributes that match the constructor.".format(f, classname)
        attr = getattr(self, f)
        if f in defaults and attr is defaults[f]:
            continue
        argstr.append("{}={}".format(f, pprint.pformat(attr)))
    return "imgaug.{}({})".format(classname, ', '.join(argstr))


ImagePlaceholder = namedtuple("ImagePlaceholder", ["shape"])


class ImageAugmentor(object):
    """
    Base class for an augmentor

    ImageAugmentor should take images of type uint8 in range [0, 255], or
    floating point images in range [0, 1] or [0, 255].

    Attributes:
        rng: a numpy :class:`RandomState`
    """

    def __init__(self):
        self.reset_state()

        # only available on Unix after Python 3.7
        if hasattr(os, 'register_at_fork'):
            os.register_at_fork(
                after_in_child=lambda: _reset_augmentor_after_fork(weakref.ref(self)))

    def _init(self, params=None):
        if params:
            for k, v in params.items():
                if k != 'self' and not k.startswith('_'):
                    setattr(self, k, v)

    def reset_state(self):
        """
        Reset rng and other state of the augmentor.

        Similar to :meth:`DataFlow.reset_state`, the caller of Augmentor
        is responsible for calling this method (once or more times) in the **process that uses the augmentor**
        before using it.

        If you use a built-in augmentation dataflow (:class:`AugmentImageComponent`, etc),
        this method will be called in the dataflow's own `reset_state` method.

        If you use Python≥3.7 on Unix, this method will be automatically called after fork,
        and you do not need to bother calling it.
        """
        self.rng = get_rng(self)

    def _rand_range(self, low=1.0, high=None, size=None):
        """
        Generate uniform float random number between low and high using `self.rng`.
        """
        if high is None:
            low, high = 0, low
        if size is None:
            size = []
        return self.rng.uniform(low, high, size).astype("float32")

    def __str__(self):
        try:
            return _default_repr(self)
        except AssertionError as e:
            log_once(e.args[0], 'warn')
            return super(Augmentor, self).__repr__()

    __repr__ = __str__

    def get_transform(self, img):
        """
        Instantiate a :class:`Transform` object to be used given the input image.
        Subclasses should implement this method.

        The :class:`ImageAugmentor` often has random policies which generate deterministic transform.
        Any of those random policies should happen inside this method and instantiate
        an actual deterministic transform to be performed.
        The returned :class:`Transform` object should perform deterministic transforms
        through its :meth:`apply_*` method.

        In this way, the returned :class:`Transform` object can be used to transform not only the
        input image, but other images or coordinates associated with the image.

        Args:
            img (ndarray): see notes of this class on the requirements.

        Returns:
            Transform
        """
        # This should be an abstract method
        # But we provide an implementation that uses the old interface,
        # for backward compatibility
        log_once("The old augmentor interface was deprecated. "
                 "Please implement {} with `get_transform` instead!".format(self.__class__.__name__),
                 "warning")

        def legacy_augment_coords(self, coords, p):
            try:
                return self._augment_coords(coords, p)
            except AttributeError:
                pass
            try:
                return self.augment_coords(coords, p)
            except AttributeError:
                pass
            return coords  # this is the old default

        p = None  # the default return value for this method
        try:
            p = self._get_augment_params(img)
        except AttributeError:
            pass
        try:
            p = self.get_augment_params(img)
        except AttributeError:
            pass

        from .transform import BaseTransform, TransformFactory
        if isinstance(p, BaseTransform):  # some old augs return Transform already
            return p

        return TransformFactory(name="LegacyConversion -- " + str(self),
                                apply_image=lambda img: self._augment(img, p),
                                apply_coords=lambda coords: legacy_augment_coords(self, coords, p))

    def augment(self, img):
        """
        Create a transform, and apply it to augment the input image.

        This can save you one line of code, when you only care the augmentation of "one image".
        It will not return the :class:`Transform` object to you
        so you won't be able to apply the same transformation on
        other data associated with the image.

        Args:
            img (ndarray): see notes of this class on the requirements.

        Returns:
            img: augmented image.
        """
        check_dtype(img)
        t = self.get_transform(img)
        return t.apply_image(img)

    # ###########################
    # Legacy interfaces:
    # ###########################
    @deprecated("Please use `get_transform` instead!", "2020-06-06", max_num_warnings=3)
    def augment_return_params(self, d):
        t = self.get_transform(d)
        return t.apply_image(d), t

    @deprecated("Please use `transform.apply_image` instead!", "2020-06-06", max_num_warnings=3)
    def augment_with_params(self, d, param):
        return param.apply_image(d)

    @deprecated("Please use `transform.apply_coords` instead!", "2020-06-06", max_num_warnings=3)
    def augment_coords(self, coords, param):
        return param.apply_coords(coords)


class AugmentorList(ImageAugmentor):
    """
    Augment an image by a list of augmentors
    """

    def __init__(self, augmentors):
        """
        Args:
            augmentors (list): list of :class:`ImageAugmentor` instance to be applied.
        """
        assert isinstance(augmentors, (list, tuple)), augmentors
        self.augmentors = augmentors
        super(AugmentorList, self).__init__()

    def reset_state(self):
        """ Will reset state of each augmentor """
        super(AugmentorList, self).reset_state()
        for a in self.augmentors:
            a.reset_state()

    def get_transform(self, img):
        check_dtype(img)
        assert img.ndim in [2, 3], img.ndim

        from .transform import LazyTransform, TransformList
        # The next augmentor requires the previous one to finish.
        # So we have to use LazyTransform
        tfms = []
        for idx, a in enumerate(self.augmentors):
            if idx == 0:
                t = a.get_transform(img)
            else:
                t = LazyTransform(a.get_transform)

            if isinstance(t, TransformList):
                tfms.extend(t.tfms)
            else:
                tfms.append(t)
        return TransformList(tfms)

    def __str__(self):
        repr_each_aug = ",\n".join(["  " + repr(x) for x in self.augmentors])
        return "imgaug.AugmentorList([\n{}])".format(repr_each_aug)

    __repr__ = __str__


Augmentor = ImageAugmentor
"""
Legacy name. Augmentor and ImageAugmentor are now the same thing.
"""


class PhotometricAugmentor(ImageAugmentor):
    """
    A base class for ImageAugmentor which only affects pixels.

    Subclass should implement `_get_params(img)` and `_impl(img, params)`.
    """
    def get_transform(self, img):
        p = self._get_augment_params(img)
        from .transform import PhotometricTransform
        return PhotometricTransform(func=lambda img: self._augment(img, p),
                                    name="from " + str(self))

    def _get_augment_params(self, _):
        return None



# -*- coding: utf-8 -*-
# File: argtools.py


import inspect
import functools

from . import logger

__all__ = ['map_arg', 'memoized', 'memoized_method', 'graph_memoized', 'shape2d', 'shape4d',
           'memoized_ignoreargs', 'log_once']


def map_arg(**maps):
    """
    Apply a mapping on certain argument before calling the original function.

    Args:
        maps (dict): {argument_name: map_func}
    """
    def deco(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # getcallargs was deprecated since 3.5
            sig = inspect.signature(func)
            argmap = sig.bind_partial(*args, **kwargs).arguments
            for k, map_func in maps.items():
                if k in argmap:
                    argmap[k] = map_func(argmap[k])
            return func(**argmap)
        return wrapper
    return deco


memoized = functools.lru_cache(maxsize=None)
""" Alias to :func:`functools.lru_cache`
WARNING: memoization will keep keys and values alive!
"""


def graph_memoized(func):
    """
    Like memoized, but keep one cache per default graph.
    """

    # TODO it keeps the graph alive
    from ..compat import tfv1
    GRAPH_ARG_NAME = '__IMPOSSIBLE_NAME_FOR_YOU__'

    @memoized
    def func_with_graph_arg(*args, **kwargs):
        kwargs.pop(GRAPH_ARG_NAME)
        return func(*args, **kwargs)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        assert GRAPH_ARG_NAME not in kwargs, "No Way!!"
        graph = tfv1.get_default_graph()
        kwargs[GRAPH_ARG_NAME] = graph
        return func_with_graph_arg(*args, **kwargs)
    return wrapper


_MEMOIZED_NOARGS = {}


def memoized_ignoreargs(func):
    """
    A decorator. It performs memoization ignoring the arguments used to call
    the function.
    """
    def wrapper(*args, **kwargs):
        if func not in _MEMOIZED_NOARGS:
            res = func(*args, **kwargs)
            _MEMOIZED_NOARGS[func] = res
            return res
        return _MEMOIZED_NOARGS[func]
    return wrapper


def shape2d(a):
    """
    Ensure a 2D shape.

    Args:
        a: a int or tuple/list of length 2

    Returns:
        list: of length 2. if ``a`` is a int, return ``[a, a]``.
    """
    if type(a) == int:
        return [a, a]
    if isinstance(a, (list, tuple)):
        assert len(a) == 2
        return list(a)
    raise RuntimeError("Illegal shape: {}".format(a))


def get_data_format(data_format, keras_mode=True):
    if keras_mode:
        dic = {'NCHW': 'channels_first', 'NHWC': 'channels_last'}
    else:
        dic = {'channels_first': 'NCHW', 'channels_last': 'NHWC'}
    ret = dic.get(data_format, data_format)
    if ret not in dic.values():
        raise ValueError("Unknown data_format: {}".format(data_format))
    return ret


def shape4d(a, data_format='NHWC'):
    """
    Ensuer a 4D shape, to use with 4D symbolic functions.

    Args:
        a: a int or tuple/list of length 2

    Returns:
        list: of length 4. if ``a`` is a int, return ``[1, a, a, 1]``
            or ``[1, 1, a, a]`` depending on data_format.
    """
    s2d = shape2d(a)
    if get_data_format(data_format, False) == 'NHWC':
        return [1] + s2d + [1]
    else:
        return [1, 1] + s2d


@memoized
def log_once(message, func='info'):
    """
    Log certain message only once. Call this function more than one times with
    the same message will result in no-op.

    Args:
        message(str): message to log
        func(str): the name of the logger method. e.g. "info", "warn", "error".
    """
    getattr(logger, func)(message)


def call_only_once(func):
    """
    Decorate a method or property of a class, so that this method can only
    be called once for every instance.
    Calling it more than once will result in exception.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self = args[0]
        # cannot use hasattr here, because hasattr tries to getattr, which
        # fails if func is a property
        assert func.__name__ in dir(self), "call_only_once can only be used on method or property!"

        if not hasattr(self, '_CALL_ONLY_ONCE_CACHE'):
            cache = self._CALL_ONLY_ONCE_CACHE = set()
        else:
            cache = self._CALL_ONLY_ONCE_CACHE

        cls = type(self)
        # cannot use ismethod(), because decorated method becomes a function
        is_method = inspect.isfunction(getattr(cls, func.__name__))
        assert func not in cache, \
            "{} {}.{} can only be called once per object!".format(
                'Method' if is_method else 'Property',
                cls.__name__, func.__name__)
        cache.add(func)

        return func(*args, **kwargs)

    return wrapper


def memoized_method(func):
    """
    A decorator that performs memoization on methods. It stores the cache on the object instance itself.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self = args[0]
        assert func.__name__ in dir(self), "memoized_method can only be used on method!"

        if not hasattr(self, '_MEMOIZED_CACHE'):
            cache = self._MEMOIZED_CACHE = {}
        else:
            cache = self._MEMOIZED_CACHE

        key = (func, ) + args[1:] + tuple(kwargs)
        ret = cache.get(key, None)
        if ret is not None:
            return ret
        value = func(*args, **kwargs)
        cache[key] = value
        return value

    return wrapper


if __name__ == '__main__':
    class A():
        def __init__(self):
            self._p = 0

        @call_only_once
        def f(self, x):
            print(x)

        @property
        def p(self):
            return self._p

        @p.setter
        @call_only_once
        def p(self, val):
            self._p = val

    a = A()
    a.f(1)

    b = A()
    b.f(2)
    b.f(1)

    print(b.p)
    print(b.p)
    b.p = 2
    print(b.p)
    b.p = 3
    print(b.p)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
from dataclasses import asdict, dataclass, field
from typing import List, Optional, Tuple, Type, Union

from loguru import logger

from transformers import HfArgumentParser
from transformers import TrainingArguments as HfTrainingArguments


def compare_arguments(one_args, another_args, return_identical=False):
    #  assert isinstance(
    #      another_args, BaseArguments
    #  ), f"another_args {type(another_args)} must be subtype of BaseArguments."

    clsname = one_args.__class__.__name__

    diff_attrs = []
    identical_attrs = []
    left_only_attrs = []
    right_only_attrs = []
    for k, v in one_args.__dict__.items():
        if k not in another_args.__dict__:
            left_only_attrs.append((clsname, k, v))
        else:
            v1 = getattr(another_args, k)
            if v == v1:
                if return_identical:
                    identical_attrs.append((clsname, k, v))
            else:
                if isinstance(v, BaseArguments):
                    comp_result = v.compare(v1, return_identical)
                    diff_attrs.extend(comp_result['diff'])
                    left_only_attrs.extend(comp_result['left_only'])
                    right_only_attrs.extend(comp_result['right_only'])
                    if return_identical:
                        identical_attrs.extend(comp_result['identical'])
                else:
                    diff_attrs.append((clsname, k, v, v1))

    for k, v in another_args.__dict__.items():
        if k not in one_args.__dict__:
            right_only_attrs.append((another_args.__class__.__name__, k, v))

    comp_result = {
        'diff': diff_attrs,
        'left_only': left_only_attrs,
        'right_only': right_only_attrs,
    }
    if return_identical:
        comp_result['identical'] = identical_attrs

    return comp_result


@dataclass
class DataArguments():
    task_name: Optional[str] = field(
        default=None,
        metadata={"help": "The name of the task to train on: "},
    )
    max_seq_length: int = field(
        default=128,
        metadata={
            "help":
            "The maximum total input sequence length after tokenization. "
            "Sequences longer than this will be truncated, sequences shorter will be padded."
        },
    )
    train_file: Optional[str] = field(
        default=None,
        metadata={"help": "The path of train file."},
    )
    test_file: Optional[str] = field(
        default=None,
        metadata={"help": "The path of test file."},
    )
    validation_file: Optional[str] = field(
        default=None,
        metadata={"help": "The path of validation file."},
    )
    pad_to_max_length: Optional[bool] = field(
        default=True, metadata={"help": "Pad to max length"})

    overwrite_cache: bool = field(
        default=False,
        metadata={"help": ("Overwrite the content of the cache directory.")},
    )
    training_samples_rate: float = field(
        default=1.0,
        metadata={"help": "How many rate of train samples to be used."},
    )

    #  split_ratio: Optional[Union[float, List, Tuple]] = field(
    #      default=0.9,
    #      #  default_factory=list,
    #      metadata={
    #          "help":
    #          "Split ratio of train/eval/test dataset. float for train or [train, eval, test]"
    #      },
    #  )

    def __post_init__(self):
        pass


@dataclass
class ModelArguments():
    model_name_or_path: str = field(
        default=None,
        metadata={"help": "Model name or path."},
    )
    config_name: Optional[str] = field(
        default=None,
        metadata={
            "help":
            "Pretrained config name or path if not the same as model_name"
        })
    tokenizer_name: Optional[str] = field(
        default=None,
        metadata={
            "help":
            "Pretrained tokenizer name or path if not the same as model_name"
        })
    cache_dir: str = field(
        default=None,
        metadata={"help": "Cache dir"},
    )
    use_fast_tokenizer: bool = field(
        default=True,
        metadata={
            "help":
            "Whether to use one of the fast tokenizer (backed by the tokenizers library) or not."
        },
    )
    model_revision: str = field(
        default="main",
        metadata={
            "help":
            "The specific model version to use (can be a branch name, tag name or commit id)."
        },
    )
    use_auth_token: bool = field(
        default=False,
        metadata={
            "help":
            "Will use the token generated when running `transformers-cli login` (necessary to use this script "
            "with private models)."
        },
    )

    def __post_init__(self):
        pass


@dataclass
class TrainingArguments(HfTrainingArguments):

    theta_version: Optional[str] = field(
        default="0.24",
        metadata={"help": "Theta version."},
    )
    #  output_dir: Optional[str] = field(
    #      default=None,
    #      metadata={
    #          "help":
    #          "The output directory where the model predictions and checkpoints will be written."
    #      },
    #  )
    #  overwrite_output_dir: bool = field(
    #      default=False,
    #      metadata={
    #          "help":
    #          ("Overwrite the content of the output directory."
    #           "Use this to continue training if output_dir points to a checkpoint directory."
    #           )
    #      },
    #  )
    #  run_name: Optional[str] = field(
    #      default=None,
    #      metadata={
    #          "help":
    #          "An optional descriptor for the run. Notably used for wandb logging."
    #      },
    #  )
    #
    #  learning_rate: float = field(
    #      default=2e-5,
    #      metadata={"help": "The initial learning rate for AdamW."},
    #  )
    #
    #  weight_decay: float = field(
    #      default=0.0,
    #      metadata={"help": "Weight decay for AdamW if we apply some."},
    #  )
    #
    #  adam_epsilon: float = field(
    #      default=1e-8,
    #      metadata={"help": "Epsilon for AdamW optimizer."},
    #  )
    #
    #  num_train_epochs: float = field(
    #      default=3.0,
    #      metadata={"help": "Total number of training epochs to perform."},
    #  )
    #  max_steps: int = field(
    #      default=-1,
    #      metadata={
    #          "help":
    #          "If > 0: set total number of training steps to perform. Override num_train_epochs."
    #      },
    #  )
    #
    #  warmup_steps: int = field(
    #      default=0,
    #      metadata={"help": "Linear warmup over warmup_steps."},
    #  )
    #
    #  lr_scheduler_type: str = field(
    #      default="linear",
    #      metadata={
    #          "help":
    #          "The scheduler type to use. One of SchedulerType "
    #          "['linear', 'cosine', 'cosine_with_restarts', 'polynomial', 'constant', 'constant_with_warmup']"
    #      },
    #  )
    #
    #  seed: int = field(
    #      default=42,
    #      metadata={
    #          "help":
    #          "Random seed that will be set at the beginning of training."
    #      },
    #  )
    #
    #  fp16: bool = field(
    #      default=False,
    #      metadata={
    #          "help":
    #          "Whether to use 16-bit (mixed) precision (through NVIDIA Apex) instead of 32-bit"
    #      },
    #  )
    #
    #  per_device_train_batch_size: int = field(
    #      default=8,
    #      metadata={"help": "Batch size per GPU/TPU core/CPU for training."},
    #  )
    #
    #  per_device_eval_batch_size: int = field(
    #      default=8,
    #      metadata={"help": "Batch size per GPU/TPU core/CPU for evaluation."},
    #  )
    #
    #  gradient_accumulation_steps: int = field(
    #      default=1,
    #      metadata={
    #          "help":
    #          "Number of updates steps to accumulate before performing a backward/update pass."
    #      },
    #  )
    #
    #  label_smoothing_factor: float = field(
    #      default=0.0,
    #      metadata={
    #          "help":
    #          "The label smoothing epsilon to apply (zero means no label smoothing)."
    #      },
    #  )
    #
    #  adafactor: bool = field(
    #      default=False,
    #      metadata={"help": "Whether or not to replace AdamW by Adafactor."},
    #  )
    #
    #  group_by_length: bool = field(
    #      default=False,
    #      metadata={
    #          "help":
    #          "Whether or not to group samples of roughly the same length together when batching."
    #      },
    #  )
    #
    #  logging_steps: int = field(
    #      default=500,
    #      metadata={"help": "Log every X updates steps."},
    #  )
    #
    #  evaluation_strategy: str = field(
    #      default="no",
    #      metadata={
    #          "help":
    #          "The evaluation strategy to use. One of EvaluationStrategy ['no', 'steps', 'epochs']."
    #      },
    #  )
    #
    #  eval_steps: int = field(
    #      default=None,
    #      metadata={"help": "Run an evaluation every X steps."},
    #  )
    #
    #  save_steps: int = field(
    #      default=500,
    #      metadata={"help": "Save checkpoint every X updates steps."},
    #  )
    #
    #  save_total_limit: Optional[int] = field(
    #      default=None,
    #      metadata={
    #          "help":
    #          ("Limit the total amount of checkpoints."
    #           "Deletes the older checkpoints in the output_dir. Default is unlimited checkpoints"
    #           )
    #      },
    #  )
    #
    #  load_best_model_at_end: Optional[bool] = field(
    #      default=False,
    #      metadata={
    #          "help":
    #          "Whether or not to load the best model found during training at the end of training."
    #      },
    #  )
    #  metric_for_best_model: Optional[str] = field(
    #      default=None,
    #      metadata={
    #          "help": "The metric to use to compare two different models."
    #      })
    #  greater_is_better: Optional[bool] = field(
    #      default=None,
    #      metadata={
    #          "help":
    #          "Whether the `metric_for_best_model` should be maximized or not."
    #      })
    #
    #  do_train: bool = field(default=False,
    #                         metadata={"help": "Whether to run training."})
    #  do_eval: bool = field(
    #      default=False,
    #      metadata={"help": "Whether to run eval on the dev set."})
    #  do_predict: bool = field(
    #      default=False,
    #      metadata={"help": "Whether to run predictions on the test set."})

    #  def __post_init__(self):
    #      pass


@dataclass
class TaskArguments():
    rootdir: str = "outputs"
    data_args: DataArguments = None
    model_args: ModelArguments = None
    training_args: TrainingArguments = None
    remaining_args: list = field(default_factory=list)

    #  data_args: DataArguments = field(
    #      default=DataArguments(),
    #      metadata={"help": "Arguments of data."},
    #  )
    #  model_args: ModelArguments = field(
    #      default=ModelArguments(),
    #      metadata={"help": "Arguments of model."},
    #  )
    #  training_args: TrainingArguments = field(
    #      default=TrainingArguments(),
    #      metadata={"help": "Arguments of training."},
    #  )
    #  remaining_args: list = field(
    #      default_factory=list,
    #      metadata={"help": "Remaining arguments come from command-line."},
    #  )

    def __post_init__(self):
        pass

    @classmethod
    def parse_args(cls):
        parser = HfArgumentParser(
            (DataArguments, ModelArguments, TrainingArguments))
        #  HfTrainingArguments))  #TrainingArguments))
        remaining_args = None
        if len(sys.argv) == 2 and sys.argv[1].endswith(".json"):
            # If we pass only one argument to the script and it's the path to a json file,
            # let's parse it to get our arguments.
            data_args, model_args, training_args = parser.parse_json_file(
                json_file=os.path.abspath(sys.argv[1]))
        else:
            data_args, model_args, training_args, remaining_args = parser.parse_args_into_dataclasses(
                return_remaining_strings=True)
        task_args = TaskArguments(data_args=data_args,
                                  model_args=model_args,
                                  training_args=training_args,
                                  remaining_args=remaining_args)
        return task_args

    @classmethod
    def merge_args(cls, args, task_args):
        from dataclasses import asdict
        for k, v in asdict(task_args).items():
            setattr(args, k, v)
        return args


if __name__ == '__main__':
    task_args = TaskArguments.parse_args()
    logger.debug(f"{task_args}")
    logger.info(f"{asdict(task_args)}")

    default_args = TaskArguments()
    comp_result = task_args.compare(default_args)
    logger.warning(f"{comp_result}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
import torch.nn as nn
from torch.nn import CrossEntropyLoss, MSELoss
from transformers.models.xlnet import (XLNetModel, XLNetPreTrainedModel)
from transformers.models.xlnet.modeling_xlnet import XLNetForSequenceClassificationOutput
from transformers.modeling_utils import SequenceSummary


class XLNetForSequenceClassification(XLNetPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.transformer = XLNetModel(config)
        self.sequence_summary = SequenceSummary(config)
        self.logits_proj = nn.Linear(config.d_model, config.num_labels)

        self.init_weights()

    #  @add_start_docstrings_to_model_forward(
    #      XLNET_INPUTS_DOCSTRING.format("batch_size, sequence_length"))
    #  @add_code_sample_docstrings(
    #      tokenizer_class=_TOKENIZER_FOR_DOC,
    #      checkpoint="xlnet-base-cased",
    #      output_type=XLNetForSequenceClassificationOutput,
    #      config_class=_CONFIG_FOR_DOC,
    #  )
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        mems=None,
        perm_mask=None,
        target_mapping=None,
        token_type_ids=None,
        input_mask=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        use_mems=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        **kwargs,  # delete when `use_cache` is removed in XLNetModel
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size,)`, `optional`):
            Labels for computing the sequence classification/regression loss. Indices should be in ``[0, ...,
            config.num_labels - 1]``. If ``config.num_labels == 1`` a regression loss is computed (Mean-Square loss),
            If ``config.num_labels > 1`` a classification loss is computed (Cross-Entropy).
        """
        #  return_dict = return_dict if return_dict is not None else self.config.use_return_dict
        return_dict = True

        transformer_outputs = self.transformer(
            input_ids,
            attention_mask=attention_mask,
            mems=mems,
            perm_mask=perm_mask,
            target_mapping=target_mapping,
            token_type_ids=token_type_ids,
            input_mask=input_mask,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            use_mems=use_mems,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            **kwargs,
        )

        output = transformer_outputs[0]

        output = self.sequence_summary(output)
        logits = self.logits_proj(output)

        loss = None
        if labels is not None:
            if self.num_labels == 1:
                #  We are doing regression
                loss_fct = MSELoss()
                loss = loss_fct(logits.view(-1), labels.view(-1))
            else:
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(logits.view(-1, self.num_labels),
                                labels.view(-1))

        #  if not return_dict:
        #  output = (logits, ) + transformer_outputs[1:]
        #  return ((loss, ) + output) if loss is not None else output

        return XLNetForSequenceClassificationOutput(
            loss=loss,
            logits=logits,
            mems=transformer_outputs.mems,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )



#  import torch
#  from torch import nn
#
#  class DiceLoss(nn.Module):
#      """DiceLoss implemented from 'Dice Loss for Data-imbalanced NLP Tasks'
#      Useful in dealing with unbalanced data
#      """
#      def __init__(self):
#          super(DiceLoss, self).__init__()
#
#      def forward(self,input, target):
#          '''
#          input: [N, C]
#          target: [N, ]
#          '''
#          prob = torch.softmax(input, dim=1)
#          prob = torch.gather(prob, dim=1, index=target.unsqueeze(1))
#          dsc_i = 1 - ((1 - prob) * prob) / ((1 - prob) * prob + 1)
#          dice_loss = dsc_i.mean()
#          return dice_loss

import torch
import torch.nn as nn
import numpy as np


def make_one_hot(vol, mask):
    '''
    :param vol: the segmentation map,[N, 1, :, :, :],N: batch size
    :param mask: can be a python list including all kinds of pixel values of your segmentation map; e.g. mask=[1,2,3]
    :return: [N, len(mask), :, :, :]
    '''

    lens = len(mask)
    shape = np.array(vol.shape)
    shape[1] = lens
    shape = tuple(shape)
    result = torch.zeros(shape)

    for idx, label in enumerate(mask):
        tmp = vol == label
        result[:, idx] = tmp

    return result


class DiceLoss(nn.Module):
    '''
    vol1,vol2: need to make one hot first
    '''
    def __init__(self, epsilon=1e-5):
        super(DiceLoss, self).__init__()
        self.epsilon = epsilon

    def forward(self, vol1, vol2):
        shape = vol1.shape
        total_loss = 0
        for i in range(shape[1]):
            top = 2 * torch.sum(torch.mul(vol1[:, i], vol2[:, i]), dtype=float)

            bottom = torch.sum(vol1[:, i], dtype=float) + torch.sum(
                vol2[:, i], dtype=float)
            bottom = torch.max(bottom, (torch.ones_like(bottom, dtype=float) *
                                        self.epsilon))  # add epsilon.
            loss_tmp = -1 * (top / bottom)
            total_loss += loss_tmp

        return total_loss / shape[1]


def dice_loss_binary(vol1, vol2):
    top = 2 * torch.sum(torch.mul(vol1, vol2), dtype=float)
    bottom = torch.sum(vol1, dtype=float) + torch.sum(vol2, dtype=float)
    bottom = torch.max(
        bottom, (torch.ones_like(bottom, dtype=float) * 1e-5))  # add epsilon.
    loss = -1 * (top / bottom)

    return loss



#!/usr/bin/env python

import numpy as np

from .base import ImageAugmentor
from .transform import Transform

__all__ = ['IAAugmentor', 'Albumentations']


class IAATransform(Transform):
    def __init__(self, aug, img_shape):
        self._init(locals())

    def apply_image(self, img):
        return self.aug.augment_image(img)

    def apply_coords(self, coords):
        import imgaug as IA
        points = [IA.Keypoint(x=x, y=y) for x, y in coords]
        points = IA.KeypointsOnImage(points, shape=self.img_shape)
        augmented = self.aug.augment_keypoints([points])[0].keypoints
        return np.asarray([[p.x, p.y] for p in augmented])


class IAAugmentor(ImageAugmentor):
    """
    Wrap an augmentor form the IAA library: https://github.com/aleju/imgaug.
    Both images and coordinates are supported.

    Note:
        1. It's NOT RECOMMENDED
           to use coordinates because the IAA library does not handle coordinates accurately.

        2. Only uint8 images are supported by the IAA library.

        3. The IAA library can only produces images of the same shape.

    Example:

    .. code-block:: python

        from imgaug import augmenters as iaa  # this is the aleju/imgaug library
        from tensorpack import imgaug  # this is not the aleju/imgaug library
        # or from dataflow import imgaug  # if you're using the standalone version of dataflow
        myaug = imgaug.IAAugmentor(
            iaa.Sequential([
                iaa.Sharpen(alpha=(0, 1), lightness=(0.75, 1.5)),
                iaa.Fliplr(0.5),
                iaa.Crop(px=(0, 100)),
            ])
    """

    def __init__(self, augmentor):
        """
        Args:
            augmentor (iaa.Augmenter):
        """
        super(IAAugmentor, self).__init__()
        self._aug = augmentor

    def get_transform(self, img):
        return IAATransform(self._aug.to_deterministic(), img.shape)


class AlbumentationsTransform(Transform):
    def __init__(self, aug, param):
        self._init(locals())

    def apply_image(self, img):
        return self.aug.apply(img, **self.param)


class Albumentations(ImageAugmentor):
    """
    Wrap an augmentor form the albumentations library: https://github.com/albu/albumentations.
    Coordinate augmentation is not supported by the library.

    Example:

    .. code-block:: python

        from tensorpack import imgaug
        # or from dataflow import imgaug  # if you're using the standalone version of dataflow
        import albumentations as AB
        myaug = imgaug.Albumentations(AB.RandomRotate90(p=1))
    """
    def __init__(self, augmentor):
        """
        Args:
            augmentor (albumentations.BasicTransform):
        """
        super(Albumentations, self).__init__()
        self._aug = augmentor

    def get_transform(self, img):
        return AlbumentationsTransform(self._aug, self._aug.get_params())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from loguru import logger
from ..common_args import get_main_args


def add_modeling_args(parser):
    #  parser.add_argument("--soft_label", action="store_true", help="Soft label for ner span")
    return parser


def get_args(
    experiment_params=None,
    special_args: list = None,
):
    return get_main_args(add_modeling_args,
                         experiment_params=experiment_params,
                         special_args=special_args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
import torch.nn as nn


# FGM
class FGM:
    def __init__(self, model: nn.Module, eps=1.):
        self.model = (model.module if hasattr(model, "module") else model)
        self.eps = eps
        self.backup = {}

    # only attack word embedding
    def attack(self, emb_name='word_embeddings'):
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                self.backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm and not torch.isnan(norm):
                    r_at = self.eps * param.grad / norm
                    param.data.add_(r_at)

    def restore(self, emb_name='word_embeddings'):
        for name, para in self.model.named_parameters():
            if para.requires_grad and emb_name in name:
                assert name in self.backup
                para.data = self.backup[name]

        self.backup = {}


# PGD
class PGD:
    def __init__(self, model, eps=1., alpha=0.3):
        self.model = (model.module if hasattr(model, "module") else model)
        self.eps = eps
        self.alpha = alpha
        self.emb_backup = {}
        self.grad_backup = {}

    def attack(self, emb_name='word_embeddings', is_first_attack=False):
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                if is_first_attack:
                    self.emb_backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm != 0 and not torch.isnan(norm):
                    r_at = self.alpha * param.grad / norm
                    param.data.add_(r_at)
                    param.data = self.project(name, param.data)

    def restore(self, emb_name='word_embeddings'):
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                assert name in self.emb_backup
                param.data = self.emb_backup[name]
        self.emb_backup = {}

    def project(self, param_name, param_data):
        r = param_data - self.emb_backup[param_name]
        if torch.norm(r) > self.eps:
            r = self.eps * r / torch.norm(r)
        return self.emb_backup[param_name] + r

    def backup_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None:
                self.grad_backup[name] = param.grad.clone()

    def restore_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None:
                param.grad = self.grad_backup[name]



# -*- coding: utf-8 -*-
# File: ilsvrc.py

import numpy as np
import os
import tarfile
import tqdm

from ...utils import logger
from ...utils.fs import download, get_dataset_path, mkdir_p
from ...utils.loadcaffe import get_caffe_pb
from ...utils.timer import timed_operation
from ..base import RNGDataFlow

__all__ = ['ILSVRCMeta', 'ILSVRC12', 'ILSVRC12Files']

CAFFE_ILSVRC12_URL = ("http://dl.caffe.berkeleyvision.org/caffe_ilsvrc12.tar.gz", 17858008)


class ILSVRCMeta(object):
    """
    Provide methods to access metadata for ILSVRC dataset.
    """

    def __init__(self, dir=None):
        if dir is None:
            dir = get_dataset_path('ilsvrc_metadata')
        self.dir = os.path.expanduser(dir)
        mkdir_p(self.dir)
        f = os.path.join(self.dir, 'synsets.txt')
        if not os.path.isfile(f):
            self._download_caffe_meta()
        self.caffepb = None

    def get_synset_words_1000(self):
        """
        Returns:
            dict: {cls_number: cls_name}
        """
        fname = os.path.join(self.dir, 'synset_words.txt')
        assert os.path.isfile(fname), fname
        lines = [x.strip() for x in open(fname).readlines()]
        return dict(enumerate(lines))

    def get_synset_1000(self):
        """
        Returns:
            dict: {cls_number: synset_id}
        """
        fname = os.path.join(self.dir, 'synsets.txt')
        assert os.path.isfile(fname)
        lines = [x.strip() for x in open(fname).readlines()]
        return dict(enumerate(lines))

    def _download_caffe_meta(self):
        fpath = download(CAFFE_ILSVRC12_URL[0], self.dir, expect_size=CAFFE_ILSVRC12_URL[1])
        tarfile.open(fpath, 'r:gz').extractall(self.dir)

    def get_image_list(self, name, dir_structure='original'):
        """
        Args:
            name (str): 'train' or 'val' or 'test'
            dir_structure (str): same as in :meth:`ILSVRC12.__init__()`.
        Returns:
            list: list of (image filename, label)
        """
        assert name in ['train', 'val', 'test']
        assert dir_structure in ['original', 'train']
        add_label_to_fname = (name != 'train' and dir_structure != 'original')
        if add_label_to_fname:
            synset = self.get_synset_1000()

        fname = os.path.join(self.dir, name + '.txt')
        assert os.path.isfile(fname), fname
        with open(fname) as f:
            ret = []
            for line in f.readlines():
                name, cls = line.strip().split()
                cls = int(cls)

                if add_label_to_fname:
                    name = os.path.join(synset[cls], name)

                ret.append((name.strip(), cls))
        assert len(ret), fname
        return ret

    def get_per_pixel_mean(self, size=None):
        """
        Args:
            size (tuple): image size in (h, w). Defaults to (256, 256).
        Returns:
            np.ndarray: per-pixel mean of shape (h, w, 3 (BGR)) in range [0, 255].
        """
        if self.caffepb is None:
            self.caffepb = get_caffe_pb()
        obj = self.caffepb.BlobProto()

        mean_file = os.path.join(self.dir, 'imagenet_mean.binaryproto')
        with open(mean_file, 'rb') as f:
            obj.ParseFromString(f.read())
        arr = np.array(obj.data).reshape((3, 256, 256)).astype('float32')
        arr = np.transpose(arr, [1, 2, 0])
        if size is not None:
            arr = cv2.resize(arr, size[::-1])
        return arr

    @staticmethod
    def guess_dir_structure(dir):
        """
        Return the directory structure of "dir".

        Args:
            dir(str): something like '/path/to/imagenet/val'

        Returns:
            either 'train' or 'original'
        """
        subdir = os.listdir(dir)[0]
        # find a subdir starting with 'n'
        if subdir.startswith('n') and \
                os.path.isdir(os.path.join(dir, subdir)):
            dir_structure = 'train'
        else:
            dir_structure = 'original'
        logger.info(
            "[ILSVRC12] Assuming directory {} has '{}' structure.".format(
                dir, dir_structure))
        return dir_structure


class ILSVRC12Files(RNGDataFlow):
    """
    Same as :class:`ILSVRC12`, but produces filenames of the images instead of nparrays.
    This could be useful when ``cv2.imread`` is a bottleneck and you want to
    decode it in smarter ways (e.g. in parallel).
    """
    def __init__(self, dir, name, meta_dir=None,
                 shuffle=None, dir_structure=None):
        """
        Same as in :class:`ILSVRC12`.
        """
        assert name in ['train', 'test', 'val'], name
        dir = os.path.expanduser(dir)
        assert os.path.isdir(dir), dir
        self.full_dir = os.path.join(dir, name)
        self.name = name
        assert os.path.isdir(self.full_dir), self.full_dir
        assert meta_dir is None or os.path.isdir(meta_dir), meta_dir
        if shuffle is None:
            shuffle = name == 'train'
        self.shuffle = shuffle

        if name == 'train':
            dir_structure = 'train'
        if dir_structure is None:
            dir_structure = ILSVRCMeta.guess_dir_structure(self.full_dir)

        meta = ILSVRCMeta(meta_dir)
        self.imglist = meta.get_image_list(name, dir_structure)

        for fname, _ in self.imglist[:10]:
            fname = os.path.join(self.full_dir, fname)
            assert os.path.isfile(fname), fname

    def __len__(self):
        return len(self.imglist)

    def __iter__(self):
        idxs = np.arange(len(self.imglist))
        if self.shuffle:
            self.rng.shuffle(idxs)
        for k in idxs:
            fname, label = self.imglist[k]
            fname = os.path.join(self.full_dir, fname)
            yield [fname, label]


class ILSVRC12(ILSVRC12Files):
    """
    Produces uint8 ILSVRC12 images of shape [h, w, 3(BGR)], and a label between [0, 999].
    The label map follows the synsets.txt file in http://dl.caffe.berkeleyvision.org/caffe_ilsvrc12.tar.gz.
    """
    def __init__(self, dir, name, meta_dir=None,
                 shuffle=None, dir_structure=None):
        """
        Args:
            dir (str): A directory containing a subdir named ``name``,
                containing the images in a structure described below.
            name (str): One of 'train' or 'val' or 'test'.
            shuffle (bool): shuffle the dataset.
                Defaults to True if name=='train'.
            dir_structure (str): One of 'original' or 'train'.
                The directory structure for the 'val' directory.
                'original' means the original decompressed directory, which only has list of image files (as below).
                If set to 'train', it expects the same two-level directory structure similar to 'dir/train/'.
                By default, it tries to automatically detect the structure.
                You probably do not need to care about this option because 'original' is what people usually have.

        Example:

        When `dir_structure=='original'`, `dir` should have the following structure:

        .. code-block:: none

            dir/
              train/
                n02134418/
                  n02134418_198.JPEG
                  ...
                ...
              val/
                ILSVRC2012_val_00000001.JPEG
                ...
              test/
                ILSVRC2012_test_00000001.JPEG
                ...

        With the downloaded ILSVRC12_img_*.tar, you can use the following
        command to build the above structure:

        .. code-block:: none

            mkdir val && tar xvf ILSVRC12_img_val.tar -C val
            mkdir test && tar xvf ILSVRC12_img_test.tar -C test
            mkdir train && tar xvf ILSVRC12_img_train.tar -C train && cd train
            find -type f -name '*.tar' | parallel -P 10 'echo {} && mkdir -p {/.} && tar xf {} -C {/.}'

        When `dir_structure=='train'`, `dir` should have the following structure:

        .. code-block:: none

            dir/
              train/
                n02134418/
                  n02134418_198.JPEG
                  ...
                ...
              val/
                n01440764/
                  ILSVRC2012_val_00000293.JPEG
                  ...
                ...
              test/
                ILSVRC2012_test_00000001.JPEG
                ...
        """
        super(ILSVRC12, self).__init__(
            dir, name, meta_dir, shuffle, dir_structure)

    """
    There are some CMYK / png images, but cv2 seems robust to them.
    https://github.com/tensorflow/models/blob/c0cd713f59cfe44fa049b3120c417cc4079c17e3/research/inception/inception/data/build_imagenet_data.py#L264-L300
    """
    def __iter__(self):
        for fname, label in super(ILSVRC12, self).__iter__():
            im = cv2.imread(fname, cv2.IMREAD_COLOR)
            assert im is not None, fname
            yield [im, label]

    @staticmethod
    def get_training_bbox(bbox_dir, imglist):
        import xml.etree.ElementTree as ET
        ret = []

        def parse_bbox(fname):
            root = ET.parse(fname).getroot()
            size = root.find('size').getchildren()
            size = map(int, [size[0].text, size[1].text])

            box = root.find('object').find('bndbox').getchildren()
            box = map(lambda x: float(x.text), box)
            return np.asarray(box, dtype='float32')

        with timed_operation('Loading Bounding Boxes ...'):
            cnt = 0
            for k in tqdm.trange(len(imglist)):
                fname = imglist[k][0]
                fname = fname[:-4] + 'xml'
                fname = os.path.join(bbox_dir, fname)
                try:
                    ret.append(parse_bbox(fname))
                    cnt += 1
                except Exception:
                    ret.append(None)
            logger.info("{}/{} images have bounding box.".format(cnt, len(imglist)))
        return ret


try:
    import cv2
except ImportError:
    from ...utils.develop import create_dummy_class
    ILSVRC12 = create_dummy_class('ILSVRC12', 'cv2')  # noqa

if __name__ == '__main__':
    meta = ILSVRCMeta()
    # print(meta.get_synset_words_1000())

    ds = ILSVRC12('/home/wyx/data/fake_ilsvrc/', 'train', shuffle=False)
    ds.reset_state()

    for _ in ds:
        from IPython import embed
        embed()
        break



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
from typing import Callable, List, Tuple, Type, Union

from loguru import logger
from tqdm import tqdm


class BaseSamples:
    def __init__(self,
                 data_file: str = None,
                 data_generator: Callable = None,
                 data_list: List = None,
                 labels_list: List = None):
        self.data_list = data_list if data_list is not None else []
        self.labels_list = labels_list if labels_list is not None else []
        self.data_generator = data_generator
        self.data_file = data_file

    def __len__(self):
        return len(self.data_list)

    def __getitem__(self, idx: int):
        return self.data_list[idx]

    def _new_instance(self, **kwargs):
        raise NotImplementedError

    def shuffle(self, random_state: int = None):
        if random_state is not None:
            random.seed(random_state)
        random.shuffle(self.data_list)

        return self

    @property
    def label2id(self):
        return {x: i for i, x in enumerate(self.labels_list)}

    @property
    def id2label(self):
        return {i: x for i, x in enumerate(self.labels_list)}

    @property
    def rows(self):
        def _rows():
            for data in self.data_list:
                yield data

        return _rows

    def load_samples(self):
        logger.warning(f"self.data_generator: {self.data_generator}")
        logger.warning(f"self.data_file: {self.data_file}")
        if self.data_generator is not None:
            self.load_samples_from_generator(self.data_generator)
        elif self.data_file is not None:
            self.load_samples_from_file(self.data_file)
        else:
            logger.warning(
                f"Unknow how to load samples. data_generator: {self.data_generator}, data_file: {self.data_file}"
            )

    def load_samples_from_generator(self, data_generator):
        self.data_list = []
        for x in data_generator():
            self.data_list.append(x)

    def load_samples_from_file(self, data_file):
        self.data_list = []
        raise NotImplementedError

    def split(self,
              ratios: Union[float, List, Tuple] = None,
              random_state=None):
        if type(ratios) == float:
            train_rate = ratios
            eval_rate = 1.0 - train_rate
            test_rate = 0.0
        elif type(ratios) == list or type(ratios) == tuple:
            if len(ratios) == 2:
                train_rate, eval_rate = ratios
                test_rate = 1.0 - train_rate - eval_rate
            elif len(ratios) == 3:
                train_rate, eval_rate, test_rate = ratios
        else:
            raise ValueError(f"ratios: {ratios} must be float or list.")

        num_total_samples = self.__len__()
        num_train_samples = int(num_total_samples * train_rate)
        if test_rate > 0.0:
            num_eval_samples = int(num_total_samples * eval_rate)
            num_test_samples = num_total_samples - num_train_samples - num_eval_samples
        else:
            num_eval_samples = num_total_samples - num_train_samples
            num_test_samples = 0

        if random_state is not None:
            from sklearn.model_selection import train_test_split
            train_data_list, eval_data_list = train_test_split(
                self.data_list,
                train_size=train_rate,
                random_state=random_state)
            train_samples = self._new_instance(data_list=train_data_list,
                                               labels_list=self.labels_list)
            eval_samples = self._new_instance(data_list=eval_data_list,
                                              labels_list=self.labels_list)

            return train_samples, eval_samples
        else:
            train_samples = self._new_instance(
                data_list=self.data_list[:num_train_samples],
                labels_list=self.labels_list)
            eval_samples = self._new_instance(
                data_list=self.data_list[num_train_samples:num_train_samples +
                                         num_eval_samples],
                labels_list=self.labels_list)
            if num_test_samples > 0:
                test_samples = self._new_instance(
                    data_list=self.data_list[-num_test_samples:],
                    labels_list=self.labels_list)
                return train_samples, eval_samples, test_samples
            else:
                return train_samples, eval_samples


def check_labels(glue_labels, labels_list):
    unknown_labels = set(labels_list) - set(glue_labels)
    if any(unknown_labels):
        logger.warning(
            f"Labels {unknown_labels} not in glue_labels: {glue_labels}")
    notexist_labels = set(glue_labels) - set(labels_list)
    if any(notexist_labels):
        logger.warning(
            f"Labels {notexist_labels} not in labels_list: {set(labels_list)}")


class GlueSamples(BaseSamples):
    def __init__(self, **kwargs):
        super(GlueSamples, self).__init__(**kwargs)

    def _new_instance(self, **kwargs):
        return GlueSamples(**kwargs)

    #  @classmethod
    #  def from_generator(cls,
    #                     data_generator: Callable,
    #                     glue_labels,
    #                     data_source: str = None):
    #      samples = GlueSamples(data_generator=data_generator,
    #                            labels_list=glue_labels)
    #      for sid, text_a, text_b, label in data_generator(data_source):
    #          samples.data_list.append((sid, text_a, text_b, label))
    #
    #      return samples


class EntitySamples(BaseSamples):
    def __init__(self, **kwargs):
        super(EntitySamples, self).__init__(**kwargs)

    def _new_instance(self, **kwargs):
        return EntitySamples(**kwargs)

    @property
    def label2id(self):
        return {x: i + 1 for i, x in enumerate(self.labels_list)}

    @property
    def id2label(self):
        return {i + 1: x for i, x in enumerate(self.labels_list)}

    #  @classmethod
    #  def from_generator(cls,
    #                     data_generator: Callable,
    #                     ner_labels,
    #                     data_source: str = None):
    #      samples = EntitySamples(data_generator=data_generator,
    #                              labels_list=ner_labels)
    #      categories = []
    #      for sid, text, _, json_tags in data_generator(data_source):
    #          """
    #          json_tag: {'category': 'Person', 'start': 0, 'mention': 'name'}
    #          """
    #          samples.data_list.append((sid, text, json_tags))
    #
    #          for tag in json_tags:
    #              c = tag['category']
    #              if c not in categories:
    #                  categories.append(c)
    #      categories = sorted(list(set(categories)))
    #      logger.info(f"Loaded categories: {categories}")
    #
    #      if not samples.labels_list:
    #          samples.labels_list = categories
    #
    #      return samples


class SPOSamples(BaseSamples):
    def __init__(self, **kwargs):
        super(SPOSamples, self).__init__(**kwargs)

    def _new_instance(self, **kwargs):
        return SPOSamples(**kwargs)

    #  @classmethod
    #  def from_generator(cls,
    #                     data_generator: Callable,
    #                     ner_labels,
    #                     data_source: str = None):
    #      samples = SPOSamples(labels_list=ner_labels)
    #      categories = []
    #      for sid, text, _, json_tags in data_generator(data_source):
    #          """
    #          json_tag: {'category': 'Person', 'start': 0, 'mention': 'name'}
    #          """
    #          samples.data_list.append((sid, text, json_tags))
    #
    #          for tag in json_tags:
    #              c = tag['category']
    #              if c not in categories:
    #                  categories.append(c)
    #      categories = sorted(list(set(categories)))
    #      logger.info(f"Loaded categories: {categories}")
    #
    #      if not samples.labels_list:
    #          samples.labels_list = categories
    #
    #      return samples



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import math
import torch
import torch.nn as nn
#  from torchcrf import CRF
from itertools import repeat


class LabelSmoothingCrossEntropy(nn.Module):
    def __init__(self, eps=0.1, reduction='mean', ignore_index=-100):
        super(LabelSmoothingCrossEntropy, self).__init__()
        self.eps = eps
        self.reduction = reduction
        self.ignore_index = ignore_index

    def forward(self, output, target):
        c = output.size()[-1]
        log_pred = torch.log_softmax(output, dim=-1)
        if self.reduction == 'sum':
            loss = -log_pred.sum()
        else:
            loss = -log_pred.sum(dim=-1)
            if self.reduction == 'mean':
                loss = loss.mean()

        return loss * self.eps / c + (1 -
                                      self.eps) * torch.nn.functional.nll_loss(
                                          log_pred,
                                          target,
                                          reduction=self.reduction,
                                          ignore_index=self.ignore_index)


class FocalLoss(nn.Module):
    """Multi-class Focal loss implementation"""
    def __init__(self,
                 gamma=2,
                 weight=None,
                 reduction='mean',
                 ignore_index=-100):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.weight = weight
        self.ignore_index = ignore_index
        self.reduction = reduction

    def forward(self, input, target):
        """
        input: [N, C]
        target: [N, ]
        """
        log_pt = torch.log_softmax(input, dim=1)
        pt = torch.exp(log_pt)
        log_pt = (1 - pt)**self.gamma * log_pt
        loss = torch.nn.functional.nll_loss(log_pt,
                                            target,
                                            self.weight,
                                            reduction=self.reduction,
                                            ignore_index=self.ignore_index)
        return loss


class SpatialDropout(nn.Module):
    """
    对字级别的向量进行丢弃
    """
    def __init__(self, drop_prob):
        super(SpatialDropout, self).__init__()
        self.drop_prob = drop_prob

    @staticmethod
    def _make_noise(input):
        return input.new().resize_(input.size(0), *repeat(1,
                                                          input.dim() - 2),
                                   input.size(2))

    def forward(self, inputs):
        output = inputs.clone()
        if not self.training or self.drop_prob == 0:
            return inputs
        else:
            noise = self._make_noise(inputs)
            if self.drop_prob == 1:
                noise.fill_(0)
            else:
                noise.bernoulli_(1 - self.drop_prob).div_(1 - self.drop_prob)
            noise = noise.expand_as(inputs)
            output.mul_(noise)
        return output


class ConditionalLayerNorm(nn.Module):
    def __init__(self, normalized_shape, cond_shape, eps=1e-12):
        super().__init__()

        self.eps = eps

        self.weight = nn.Parameter(torch.Tensor(normalized_shape))
        self.bias = nn.Parameter(torch.Tensor(normalized_shape))

        self.weight_dense = nn.Linear(cond_shape, normalized_shape, bias=False)
        self.bias_dense = nn.Linear(cond_shape, normalized_shape, bias=False)

        self.reset_weight_and_bias()

    def reset_weight_and_bias(self):
        """
        此处初始化的作用是在训练开始阶段不让 conditional layer norm 起作用
        """
        nn.init.ones_(self.weight)
        nn.init.zeros_(self.bias)

        nn.init.zeros_(self.weight_dense.weight)
        nn.init.zeros_(self.bias_dense.weight)

    def forward(self, inputs, cond=None):
        assert cond is not None, 'Conditional tensor need to input when use conditional layer norm'
        cond = torch.unsqueeze(cond, 1)  # (b, 1, h*2)

        weight = self.weight_dense(cond) + self.weight  # (b, 1, h)
        bias = self.bias_dense(cond) + self.bias  # (b, 1, h)

        mean = torch.mean(inputs, dim=-1, keepdim=True)  # （b, s, 1）
        outputs = inputs - mean  # (b, s, h)

        variance = torch.mean(outputs**2, dim=-1, keepdim=True)
        std = torch.sqrt(variance + self.eps)  # (b, s, 1)

        outputs = outputs / std  # (b, s, h)

        outputs = outputs * weight + bias

        return outputs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .base_extractor import BaseExtractor


class PointerSequenceExtractor(BaseExtractor):
    def __init__(self):
        super(PointerSequenceExtractor, self).__init__()

    def decode(self, **kwarg):
        start_logits = kwarg['start_logits']
        end_logits = kwarg['end_logits']
        lens = kwarg['lens']
        if confidence in kwarg:
            confidence = kwarg['confidence']
        if overlap in kwarg:
            overlap = kwarg['overlap']
        else:
            overlap = False

        num_tokens = lens[0]
        S = []

        starts = torch.argmax(start_logits, -1).cpu().numpy()[0]  #[1:-1]
        ends = torch.argmax(end_logits, -1).cpu().numpy()[0]  #[1:-1]

        # start_logits.shape: (1, sentence_length, num_labels)
        #  logger.debug(f"start_logits: {start_logits.shape}")
        #  start_max_probs = [
        #      f"{start_logits[0][i][x]:.4f}" for i, x in enumerate(starts)
        #  ]
        #  logger.debug(f"start_max_probs: {start_max_probs}")
        #  logger.debug(f"starts: {starts.shape} {starts}")
        #  starts = [
        #      x if start_logits[0][i][x] >= confidence else 0
        #      for i, x in enumerate(starts)
        #  ]
        #  ends = [
        #      x if end_logits[0][i][x] >= confidence else 0
        #      for i, x in enumerate(ends)
        #  ]

        #  starts = [starts[0]] + [ x if x != starts[i] else 0 for i, x in enumerate(starts[1:])]
        #  ends = [ends[0]] + [ x if x != ends[i] else 0 for i, x in enumerate(ends[1:])]
        def filter_process(starts):
            new_starts = []
            for i, x in enumerate(starts):
                is_dup = False
                if i < len(starts) - 1 and x == starts[i + 1]:
                    is_dup = True
                elif i > 0 and starts[i - 1] == x:
                    is_dup = True
                if is_dup:
                    new_starts.append(0)
                else:
                    new_starts.append(x)
            return new_starts

        #  starts = filter_process(starts)
        #  ends = filter_process(ends)

        #  starts = np.array([x for x in starts if x >= 0 and x < num_tokens])
        #  ends = np.array([x for x in ends if x >= 0 and x < num_tokens])
        starts = starts[:num_tokens]
        ends = ends[:num_tokens]

        #  logger.info(f"start_pred: {starts}")
        #  logger.info(f"end_pred: {ends}")
        #  for i, s_l in enumerate(starts):
        #      if s_l == 0:
        #          continue
        #      for j, e_l in enumerate(ends[i:]):
        #          if s_l == e_l:
        #              S.append((s_l, i, i + j))
        #              break
        #          if i + j < len(starts) - 1 and starts[i + j + 1] != 0:
        #              break
        #  for i in range(len(starts) - 1):
        last_j = -1
        for i in range(len(starts)):
            if i <= last_j:
                continue
            s_l = starts[i]
            if s_l == 0:
                continue
            for j, e_l in enumerate(ends[i:]):
                if s_l == e_l:
                    if not overlap:
                        #  if sum(starts[i + 1:i + j + 1]) == 0:
                        S.append((int(s_l), i, i + j))
                        last_j = j
                        i = j + 1
                        break
                    else:
                        if sum(starts[i + 1:i + j + 1]) != 0:
                            break
                        S.append((int(s_l), i, i + j))
                if i + j < len(starts) - 1 and starts[i + j + 1] != 0:
                    break
        #  S = [x for x in S if x[1] <= x[2]]

        #  for x in S:
        #      assert x[1] >= 0 and x[2] >= 0 and x[1] <= x[2], f"S: {S}"

        return S



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy, random
import json
import os
from pathlib import Path
from collections import defaultdict

import numpy as np
import torch
from loguru import logger
from torch.autograd import Variable
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
from tqdm import tqdm

from ..utils import init_theta
from ..utils.multiprocesses import (barrier_leader_process,
                                    barrier_member_processes,
                                    is_master_process, is_multi_processes)
from ..utils.progbar import Progbar

try:
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    from tensorboardX import SummaryWriter


def get_default_optimizer_parameters(model, weight_decay):
    param_optimizers = list(model.named_parameters())
    no_decay = ['bias', 'LayerNorm.weights']
    optimizer_grouped_parameters = [{
        'params': [
            p for n, p in param_optimizers
            if not any(nd in n for nd in no_decay)
        ],
        'weight_decay':
        weight_decay
    }, {
        'params':
        [p for n, p in param_optimizers if any(nd in n for nd in no_decay)],
        'weight_decay':
        0.0
    }]

    return optimizer_grouped_parameters


enttype_labels = defaultdict(list)


def augment_examples(examples, augs=0):
    for a in range(augs):
        aug_examples = copy.deepcopy(examples)
        for e in tqdm(aug_examples, desc="Aug examples"):
            p0 = 0
            text = copy.deepcopy(e.text_a)
            new_text = ""
            new_labels = []
            for i, (label, start, end) in enumerate(e.labels):
                new_text += text[p0:start]
                if random.randint(0, 1000 - 1) < 666:
                    new_start = len(new_text)
                    new_mention = text[start:end + 1]
                    new_text += new_mention
                    new_labels.append(
                        (label, new_start, new_start + len(new_mention) - 1))
                    p0 = end + 1
                else:
                    new_start = len(new_text)
                    new_idx = random.randint(0, len(enttype_labels[label]) - 1)
                    new_mention = enttype_labels[label][new_idx]
                    new_text += new_mention
                    new_end = new_start + len(new_mention) - 1
                    new_labels.append((label, new_start, new_end))
                    p0 = end + 1
            new_text += text[p0:]
            #  logger.debug(
            #      f"text: {text} | {[ text[start:end+1] for _, start, end in e.labels]}"
            #  )
            #  logger.warning(
            #      f"new_text: {new_text} | {[ new_text[start:end+1] for _, start, end in new_labels]}"
            #  )
            e.text_a = new_text
            e.labels = new_labels

        examples.extend(aug_examples)

    return examples


def generate_dataloader(args,
                        dataset,
                        batch_size,
                        keep_order=True,
                        collate_fn=None):

    Sampler = SequentialSampler if keep_order else RandomSampler
    sampler = DistributedSampler(dataset) if is_multi_processes(
        args) else Sampler(dataset)
    dataloader = DataLoader(
        dataset,
        sampler=sampler,
        batch_size=batch_size,
        #  collate_fn=None)
        collate_fn=collate_fn)
    return dataloader


def common_batch_encode(texts, label2id, tokenizer, max_seq_length):
    all_encodes = tokenizer.batch_encode(texts, add_special_tokens=True)
    all_tokens = all_encodes['tokens']
    all_token_offsets = all_encodes['offsets']
    all_token2char = all_encodes['token2char']
    all_char2token = all_encodes['char2token']

    all_input_ids = all_encodes['ids']
    all_attention_mask = all_encodes['attention_mask']
    all_token_type_ids = all_encodes['type_ids']

    all_input_lens = [len(tokens) for tokens in all_tokens]

    all_padding_lens = [max_seq_length - n for n in all_input_lens]
    for i, (input_ids, attention_mask, token_type_ids, token2char,
            token_offsets, padding_length) in enumerate(
                tqdm(zip(all_input_ids, all_attention_mask, all_token_type_ids,
                         all_token2char, all_token_offsets, all_padding_lens),
                     desc="common_batch_encode")):

        if padding_length < 0:
            all_tokens[i] = all_tokens[i][:max_seq_length - 1] + ['[SEP]']
            all_input_ids[i] = input_ids[:max_seq_length - 1] + [102]
            all_attention_mask[i] = attention_mask[:max_seq_length]
            all_token_type_ids[i] = token_type_ids[:max_seq_length]
            all_token2char[i] = token2char[:max_seq_length]
            all_token_offsets[i] = token_offsets[:max_seq_length - 1] + [(0, 0)
                                                                         ]
            #  logger.warning(f"all_tokens[i]: {all_tokens[i]}")
            #  logger.debug(f"all_input_ids[i]: {all_input_ids[i]}")
            #  logger.debug(f"all_token_offsets[i]: {all_token_offsets[i]}")

        else:
            all_input_ids[i] = input_ids + [0] * padding_length
            all_attention_mask[i] = attention_mask + [0] * padding_length
            all_token_type_ids[i] = token_type_ids + [0] * padding_length
            all_token2char[i] = token2char + [0] * padding_length
            all_token_offsets[i] = token_offsets + [(0, 0)] * padding_length
        all_input_lens[i] = len(input_ids)

    #  logger.warning(f"all_input_ids: {all_input_ids}")
    all_input_ids = torch.from_numpy(np.array(all_input_ids, dtype=np.int64))
    all_attention_mask = torch.from_numpy(
        np.array(all_attention_mask, dtype=np.int64))
    all_token_type_ids = torch.from_numpy(
        np.array(all_token_type_ids, dtype=np.int64))
    all_input_lens = torch.from_numpy(np.array(all_input_lens, dtype=np.int64))
    all_token_offsets = torch.from_numpy(
        np.array(all_token_offsets, dtype=np.int64))

    return (all_tokens, all_token2char, all_char2token, all_input_ids,
            all_attention_mask, all_token_type_ids, all_input_lens,
            all_token_offsets)


def common_to_tensors(all_input_ids, all_attention_mask, all_token_type_ids,
                      all_input_lens, all_token_offsets):
    all_input_ids = torch.from_numpy(np.array(all_input_ids, dtype=np.int64))
    all_attention_mask = torch.from_numpy(
        np.array(all_attention_mask, dtype=np.int64))
    all_token_type_ids = torch.from_numpy(
        np.array(all_token_type_ids, dtype=np.int64))
    all_input_lens = torch.from_numpy(np.array(all_input_lens, dtype=np.int64))
    all_token_offsets = torch.from_numpy(
        np.array(all_token_offsets, dtype=np.int64))

    return all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets


class Trainer:
    def __init__(self, args):
        init_theta(args)

        self.args = args
        self.collate_fn = None

        self.eval_loss = 0.0

    #  def build_model(self, args):
    #      raise NotImplementedError

    def batch_to_inputs(self, args, batch, known_labels=True):
        raise NotImplementedError

    #  def examples_to_dataset(self, examples, max_seq_length):
    #      raise NotImplementedError

    def encode_examples(self,
                        examples,
                        max_seq_length,
                        enttype_labels=None,
                        epoch=-1):
        raise NotImplementedError

    #  def generate_dataloader(self, args, dataset, batch_size, keep_order=True):
    #
    #      return generate_dataloader(args, dataset, batch_size, keep_order)

    def on_train_step(self, args, model, step, batch):

        inputs = self.batch_to_inputs(args, batch)
        outputs = model(**inputs)
        return outputs

    def on_eval_start(self, args, eval_dataset):
        pass

    #  def on_eval_step(self, args, eval_dataset, step, model, inputs, outputs):
    #      results = {}
    #      return results

    #  def on_eval_step(self, args, model, step, batch, batch_features):
    def on_eval_step(self, args, model, step, batch):
        inputs = self.batch_to_inputs(args, batch)
        outputs = model(**inputs)
        return outputs, {}

    def on_eval_end(self, args, eval_examples):
        results = {}
        return results

    def on_predict_start(self, args, test_dataset):
        pass

    #  def on_predict_step(self, args, test_dataset, step, model, inputs,
    #                      outputs):
    def on_predict_step(self, args, model, step, batch):
        pass

    def on_predict_end(self, args, test_dataset):
        return []

    def save_args(self, args, model_path):
        from .utils import save_args
        save_args(args, model_path)

    def save_params_file(self, args, model_path):
        latest_dir = args.latest_dir
        dataset_name = args.dataset_name
        cmd_cp_params = f"cp {dataset_name}_params.py {model_path}/{dataset_name}_params.py"
        logger.warning(f"{cmd_cp_params}")
        os.system(cmd_cp_params)

    def save_model(self, args, model, tokenizer, optimizer, scheduler,
                   model_path):
        if not os.path.exists(model_path):
            os.makedirs(model_path)

        logger.warning(f"Saving model checkpoint to {model_path}")

        model_to_save = (model.module if hasattr(model, "module") else model)
        model_to_save.save_pretrained(model_path)

        tokenizer.save_vocabulary(Path(model_path).as_posix() + '/')

        self.save_args(args, model_path)

        #  torch.save(optimizer.state_dict(),
        #             os.path.join(model_path, "optimizer.pt"))
        #  torch.save(scheduler.state_dict(),
        #             os.path.join(model_path, "scheduler.pt"))

    def train(self, args, train_examples, eval_examples):

        logger.info(
            f"Start train: {len(train_examples)} train examples, {len(eval_examples)} eval examples."
        )

        self.save_args(args, args.latest_dir)
        self.save_params_file(args, args.latest_dir)

        #  if is_master_process(args):
        #      tb_writer = SummaryWriter()

        #  _, all_features = self.examples_to_dataset(train_examples,
        #                                             args.train_max_seq_length)

        for e in tqdm(train_examples + eval_examples,
                      desc="Build enttype_labels"):
            text = e.text_a
            for label, start, end in e.labels:
                mention = text[start:end + 1]
                enttype_labels[label].append(mention)
        for k, labels in enttype_labels.items():
            enttype_labels[k] = list(set(enttype_labels[k]))

        # FIXME
        train_examples = augment_examples(train_examples,
                                          augs=args.num_augments)
        all_features = self.encode_examples(train_examples,
                                            args.train_max_seq_length)

        logger.warning(f"Generate dataloader...")
        train_dataloader = generate_dataloader(
            args,
            #  train_dataset,
            all_features,
            batch_size=args.per_gpu_train_batch_size,
            keep_order=False,
            collate_fn=self.collate_fn)
        logger.info(f"Start training ...")
        logger.info(f"  Num examples    = {len(train_examples)}")
        logger.info(f"  Num epoch steps = {len(train_dataloader)}")
        logger.info(f"  Num epochs = {args.num_train_epochs}")
        logger.info(f"  Batch size = {args.per_gpu_train_batch_size}")

        steps_per_epoch = len(
            train_dataloader) // args.gradient_accumulation_steps

        if args.max_steps > 0:
            total_steps = args.max_steps
            args.num_train_epochs = args.max_steps // steps_per_epoch + 1
        else:
            total_steps = steps_per_epoch * args.num_train_epochs
        args.total_steps = total_steps

        logger.info(
            f"  Gradient Accumulation steps = {args.gradient_accumulation_steps}"
        )
        logger.info(f"  Total optimization steps = {total_steps}")

        model, optimizer, scheduler = self.build_model(args)
        tokenizer = self.tokenizer

        # Check if saved optimizer or scheduler states exist
        #  model_path = Path(args.model_path)
        #  optimizer_saved_file = model_path / "optimizer.pt"
        #  scheduler_saved_file = model_path / "scheduler.pt"
        #  if optimizer_saved_file.exists() and scheduler_saved_file.exists():
        #      optimizer.load_state_dict(torch.load(optimizer_saved_file))
        #      scheduler.load_state_dict(torch.load(scheduler_saved_file))

        model = model.to(args.device)
        #  logger.warning(
        #      f"model.named_parameters(): {list(model.named_parameters())}")
        if args.fp16:
            try:
                from apex import amp
            except ImportError:
                raise ImportError(
                    "Please install apex from https://github.com/nvidia/apex to use fp16."
                )
            model, optimizer = amp.initialize(model,
                                              optimizer,
                                              opt_level=args.fp16_opt_level)
            args.amp = amp
        # multi-gpu training (should be after apex fp16 initialization)
        if args.n_gpu > 1:
            model = torch.nn.DataParallel(model)
        # Distributed training (should be after apex fp16 initialization)
        #  if args.local_rank != -1:
        if is_multi_processes(args):
            model = torch.nn.parallel.DistributedDataParallel(
                model,
                device_ids=[args.local_rank],
                output_device=args.local_rank,
                find_unused_parameters=True)

        trained_steps = 0
        epochs_trained = 0
        steps_trained_in_current_epoch = 0
        # Check if continuing training from a checkpoint
        model_path = Path(args.model_path)
        output_dir = Path(args.output_dir)

        #  if model_path.exists() and "checkpoint" in str(model_path):
        #      # set trained_steps to trained_steps of last saved checkpoint from model path
        #      trained_steps = int(model_path.parts()[-1].split("-")[-1])
        #      epochs_trained = trained_steps // (
        #          len(train_examples) // args.gradient_accumulation_steps)
        #      steps_trained_in_current_epoch = trained_steps % (
        #          len(train_examples) // args.gradient_accumulation_steps)
        #
        #      logger.info(
        #          "  Continuing training from checkpoint, will skip to saved trained_steps"
        #      )
        #      logger.info("  Continuing training from epoch %d", epochs_trained)
        #      logger.info("  Continuing training from global step %d",
        #                  trained_steps)
        #      logger.info("  Will skip the first %d steps in the first epoch",
        #                  steps_trained_in_current_epoch)
        #
        train_loss, logging_loss = 0.0, 0.0

        best_index = args.best_index
        if best_index in ['loss']:
            best_value = float('inf')
            best_type = 'min'
        else:
            best_value = 0.0
            best_type = 'max'  # ['max', 'min']

        model.zero_grad()

        # https://arxiv.org/abs/2002.10345
        # https://github.com/lonePatient/BERT-SDA
        enable_kd = args.enable_kd
        if enable_kd:
            logger.warning("Enable knowledge distillation.")
            from torch.nn import MSELoss
            kd_loss_fct = MSELoss()
            kd_model = copy.deepcopy(model)
            kd_model.eval()

        enable_sda = args.enable_sda
        if enable_sda:
            from torch.nn import MSELoss
            sda_loss_fct = MSELoss()
            history_logits = []
            sda_teachers = args.sda_teachers
            sda_stategy = args.sda_stategy

            if args.sda_empty_first:
                teacher_models = []
            else:
                t_model = copy.deepcopy(model)
                t_model.eval()
                teacher_models = [t_model]

            #  best_logits = []

        #  train_iterator = trange(
        #      epochs_trained,
        #      int(args.num_train_epochs),
        #      desc="Epoch",
        #      disable=args.local_rank not in [-1, 0],
        #  )
        #  for epoch in train_iterator:
        for epoch in range(epochs_trained, args.num_train_epochs):
            args.epoch = epoch

            #  all_features = self.encode_examples(train_examples,
            #                                      args.train_max_seq_length,
            #                                      enttype_labels=enttype_labels,
            #                                      epoch=epoch)
            #
            #  logger.warning(f"Generate dataloader...")
            #  train_dataloader = generate_dataloader(
            #      args,
            #      #  train_dataset,
            #      all_features,
            #      batch_size=args.per_gpu_train_batch_size,
            #      keep_order=False,
            #      collate_fn=self.collate_fn)
            #  logger.info(f"Start training ...")
            #  logger.info(f"  Num examples    = {len(train_examples)}")
            #  logger.info(f"  Num epoch steps = {len(train_dataloader)}")
            #  logger.info(f"  Num epochs = {args.num_train_epochs}")
            #  logger.info(f"  Batch size = {args.per_gpu_train_batch_size}")
            #
            pbar = Progbar(target=len(train_dataloader),
                           stateful_metrics=['loss'],
                           desc=f"Epoch({epoch+1}/{args.num_train_epochs})")
            for step, batch in enumerate(train_dataloader):
                if steps_trained_in_current_epoch > 0:
                    steps_trained_in_current_epoch -= 1
                    continue

                model.train()
                batch = tuple(
                    t.to(args.device) if isinstance(t, torch.Tensor) else t
                    for t in batch)

                outputs = self.on_train_step(args, model, step, batch)
                #  inputs = self.batch_to_inputs(args, batch)
                #  outputs = model(**inputs)

                #  logger.debug(f"outputs: {outputs}")
                # -------- loss --------
                loss, logits = outputs[:2]

                #  loss += torch.tensor(self.eval_loss).cuda()

                if enable_kd:
                    inputs = self.batch_to_inputs(args, batch)
                    if "labels" in inputs:
                        inputs['labels'] = None
                    with torch.no_grad():
                        #  kd_logits = kd_model(**inputs)[0]
                        kd_logits = kd_model(**inputs)[1]
                    kd_loss = kd_loss_fct(outputs[1], kd_logits)
                    loss += args.kd_coeff * kd_loss

                if enable_sda:
                    if teacher_models:
                        inputs = self.batch_to_inputs(args, batch)
                        if "labels" in inputs:
                            inputs['labels'] = None
                        with torch.no_grad():
                            teacher_logits = [
                                m(**inputs)[1] for m in teacher_models
                            ]
                        teacher_logits = torch.stack(teacher_logits)
                        teacher_logits = torch.mean(teacher_logits, dim=0)
                        sda_loss = sda_loss_fct(logits, teacher_logits)
                        #  sda_loss = Variable(sda_loss, requires_grad=True)
                        loss += sda_loss * args.sda_coeff

                    #  if best_logits:
                    #      sda_logits = torch.stack(best_logits)
                    #      sda_logits = torch.mean(sda_logits, dim=0)
                    #      if sda_logits.shape[0] == logits.shape[0]:
                    #          sda_loss = sda_loss_fct(logits, sda_logits)
                    #          #  sda_loss = Variable(sda_loss, requires_grad=True)
                    #          loss += sda_loss * args.sda_coeff

                #  loss = Variable(loss, requires_grad=True)
                #  inputs = self.batch_to_inputs(args, batch)
                #  logger.debug(f"inputs: {inputs}")
                #  logger.info(f"loss: {loss}")

                if loss is None:
                    continue

                if args.n_gpu > 1:
                    loss = loss.mean()
                if args.gradient_accumulation_steps > 1:
                    loss = loss / args.gradient_accumulation_steps
                #  logger.info(f"loss: {loss}")

                if args.fp16:
                    with amp.scale_loss(loss, optimizer) as scaled_loss:
                        scaled_loss.backward()
                else:
                    loss.backward()

                lr = scheduler.get_last_lr()[0]
                #  lr = scheduler.get_lr()[0]
                pbar.update(step + 1,
                            values=[('lr', lr), ('loss', loss.item()),
                                    ('eval_loss', self.eval_loss)])
                train_loss += loss.item()

                if (step + 1) % args.gradient_accumulation_steps == 0:
                    if args.fp16:
                        torch.nn.utils.clip_grad_norm_(
                            amp.master_params(optimizer), args.max_grad_norm)
                    else:
                        torch.nn.utils.clip_grad_norm_(model.parameters(),
                                                       args.max_grad_norm)
                    optimizer.step()
                    scheduler.step()
                    model.zero_grad()
                    trained_steps += 1

                    if enable_kd:
                        decay = min(args.kd_decay,
                                    (1 + trained_steps) / (10 + trained_steps))
                        one_minus_decay = 1.0 - decay
                        with torch.no_grad():
                            parameters = [
                                p for p in model.parameters()
                                if p.requires_grad
                            ]
                            for s_param, param in zip(kd_model.parameters(),
                                                      parameters):
                                s_param.sub_(one_minus_decay *
                                             (s_param - param))
                    if enable_sda:
                        decay = min(args.sda_decay,
                                    (1 + trained_steps) / (10 + trained_steps))
                        one_minus_decay = 1.0 - decay
                        with torch.no_grad():
                            for sda_model in teacher_models:
                                parameters = [
                                    p for p in model.parameters()
                                    if p.requires_grad
                                ]
                                for s_param, param in zip(
                                        sda_model.parameters(), parameters):
                                    s_param.sub_(one_minus_decay *
                                                 (s_param - param))

                # -------- Save models --------
                if is_master_process(
                        args) and trained_steps % steps_per_epoch == 0:

                    # -------- Save checkpoint --------
                    if args.save_checkpoints:
                        checkpoint_dir = f"checkpoint-{trained_steps}"

                        #  checkpoint_path = output_dir / checkpoint_dir
                        checkpoint_path = Path(
                            args.latest_dir) / checkpoint_dir

                        self.save_model(args, model, tokenizer, optimizer,
                                        scheduler, checkpoint_path)

                    # -------- Evaluate --------
                    if not args.no_eval_on_each_epoch:
                        logger.info(
                            f"Epoch({epoch+1}/{args.num_train_epochs}) evaluating."
                        )
                        eval_logs = {}
                        eval_results = self.evaluate(args, model,
                                                     eval_examples)
                        for key, value in eval_results.items():
                            eval_key = "eval_{}".format(key)
                            eval_logs[eval_key] = f"{value:.6f}"
                        loss_scalar = (train_loss -
                                       logging_loss) / steps_per_epoch
                        learning_rate_scalar = scheduler.get_last_lr()[0]
                        #  learning_rate_scalar = scheduler.get_lr()[0]
                        eval_logs[
                            "learning_rate"] = f"{learning_rate_scalar:.6f}"
                        eval_logs["loss_scalar"] = f"{loss_scalar:.6f}"
                        #  for key, value in eval_logs.items():
                        #      tb_writer.add_scalar(key, value, trained_steps)
                        logger.debug(
                            json.dumps({
                                **eval_logs,
                                **{
                                    "step": trained_steps
                                }
                            }))
                        logging_loss = train_loss

                        # -------- Save best model --------
                        #  best_index = 'f1'  # ['f1', 'acc', 'recall', 'loss']
                        best_index = args.best_index
                        eval_value = eval_results[best_index]
                        #  logger.warning(
                        #      f"best_index: {best_index}, best_value: {best_value:.6f}, eval_value: {eval_value:.6f}"
                        #  )
                        is_best = False
                        if best_index in ['loss']:
                            if eval_value < best_value:
                                is_best = True
                        else:
                            if eval_value > best_value:
                                is_best = True
                        if is_best:
                            logger.warning(
                                f"Best {best_index}: {eval_value:.6f} ({eval_value - best_value:.6f})"
                            )
                            #  if enable_sda:
                            #      best_logits.append(logits.detach())
                            #      if len(best_logits) > sda_teachers:
                            #          best_logits = best_logits[1:]

                            best_value = eval_value

                            #  bestmodel_path = output_dir / f"best_fold{args.fold}"
                            bestmodel_path = Path(args.latest_dir) / "best"

                            self.save_model(args, model, tokenizer, optimizer,
                                            scheduler, bestmodel_path)

                            if args.save_checkpoints:
                                best_symlink = output_dir / "best/best_checkpoint"
                                if best_symlink.is_symlink():
                                    best_symlink.unlink()
                                os.symlink(f"../{checkpoint_dir}",
                                           best_symlink)
                        else:
                            logger.info(
                                f"dev-{best_index}/best-{best_index}: {eval_value:.6f}/{best_value:.6f}"
                            )

            #  if enable_sda:
            #      best_logits.append(logits.detach())
            #      if len(best_logits) > sda_teachers:
            #          best_logits = best_logits[1:]

            if enable_sda:
                if sda_stategy == "recent_models":
                    if len(teacher_models) >= sda_teachers:
                        teacher_models = teacher_models[1:sda_teachers + 1]
                    t_model = copy.deepcopy(model)
                    t_model.eval()
                    teacher_models.append(t_model)
                elif sda_stategy == "earliest_models":
                    if len(teacher_models) < sda_teachers:
                        t_model = copy.deepcopy(model)
                        t_model.eval()
                        teacher_models.append(t_model)
                elif sda_stategy == 'latest_model':
                    t_model = copy.deepcopy(model)
                    t_model.eval()
                    teacher_models = [t_model]
                elif sda_stategy == 'clone_models':
                    if len(teacher_models) == 0:
                        t_model = copy.deepcopy(model)
                    else:
                        t_model = copy.deepcopy(teacher_models[-1])
                    t_model.eval()
                    teacher_models.append(t_model)
                    if len(teacher_models) > sda_teachers:
                        teacher_models = teacher_models[1:sda_teachers + 1]

                #  if len(teacher_models) < sda_teachers:
                #      if teacher_models:
                #          tmodel = copy.deepcopy(teacher_models[-1])
                #      else:
                #          t_model = copy.deepcopy(model)
                #      t_model.eval()
                #      teacher_models.append(t_model)

            print(" ")
            if 'cuda' in str(args.device):
                torch.cuda.empty_cache()

        #      if args.max_steps > 0 and trained_steps > args.max_steps:
        #          epoch_iterator.close()
        #          break
        #  if args.max_steps > 0 and trained_steps > args.max_steps:
        #      train_iterator.close()
        #      break

        #  if is_multi_processes(args):
        #      tb_writer.close()
        return trained_steps, train_loss / trained_steps

    def evaluate(self, args, model, eval_examples):
        #  _, eval_features = self.examples_to_dataset(eval_examples,
        #                                              args.eval_max_seq_length)
        eval_features = self.encode_examples(eval_examples,
                                             args.eval_max_seq_length)

        self.on_eval_start(args, eval_features)
        eval_dataloader = generate_dataloader(
            args,
            #  eval_dataset,
            eval_features,
            batch_size=args.per_gpu_eval_batch_size,
            keep_order=True,
            collate_fn=self.collate_fn)

        eval_output_dir = args.output_dir
        if is_multi_processes(args) and not os.path.exists(eval_output_dir):
            os.makedirs(eval_output_dir)

        logger.info(f"Start evaluating ...")
        logger.info(f"  Num examples    = {len(eval_examples)}")
        logger.info(f"  Num epoch steps = {len(eval_dataloader)}")
        logger.info(f"  Batch size = {args.per_gpu_eval_batch_size}")

        eval_steps = 0
        eval_loss = 0.0
        pbar = Progbar(target=len(eval_dataloader),
                       stateful_metrics=['loss', 'acc', 'recall', 'f1'],
                       desc=f"Evaluating")
        for step, batch in enumerate(eval_dataloader):
            model.eval()
            batch = tuple(
                t.to(args.device) if isinstance(t, torch.Tensor) else t
                for t in batch)
            with torch.no_grad():
                #  inputs = self.batch_to_inputs(args, batch)
                #  outputs = model(**inputs)
                outputs, results = self.on_eval_step(args, model, step, batch)
                #  batch_features = eval_features[args.per_gpu_eval_batch_size *
                #                                 step:args.
                #                                 per_gpu_eval_batch_size *
                #                                 (step + 1)]
                #  outputs, results = self.on_eval_step(args, model, step, batch,
                #                                       batch_features)

                # -------- loss --------
                loss = outputs[0]

                if args.n_gpu > 1:
                    loss = loss.mean()
                eval_loss += loss.item()

            eval_steps += 1

            #  results = self.on_eval_step(args, eval_dataset, step, model,
            #                              inputs, outputs)
            values = []
            if results:
                values = [(k, v) for k, v in results.items()]
            values += [('loss', loss.item())]
            pbar.update(step + 1, values=values)

        results = self.on_eval_end(args, eval_features)

        self.eval_loss = eval_loss / eval_steps
        results['loss'] = self.eval_loss
        logger.info(f" dev: loss = {self.eval_loss:.6f}")

        return results

    def predict(self, args, model, test_examples):

        pred_output_dir = Path(args.output_dir)
        if is_multi_processes(args) and not pred_output_dir.exists():
            os.makedirs(pred_output_dir)

        #  test_dataset, _ = self.examples_to_dataset(test_examples,
        #                                             args.eval_max_seq_length)
        test_features = self.encode_examples(test_examples,
                                             args.eval_max_seq_length)
        test_dataloader = generate_dataloader(
            args,
            test_features,
            batch_size=args.per_gpu_predict_batch_size,
            keep_order=True,
            collate_fn=self.collate_fn)

        logger.info(f"Start predicting ...")
        logger.info(f"  Num examples    = {len(test_examples)}")
        logger.info(f"  Num epoch steps = {len(test_dataloader)}")
        logger.info(f"  Batch size = {args.per_gpu_predict_batch_size}")

        pbar = Progbar(target=len(test_dataloader), desc=f"Predicting")
        self.on_predict_start(args, test_features)
        for step, batch in enumerate(test_dataloader):
            model.eval()
            batch = tuple(
                t.to(args.device) if isinstance(t, torch.Tensor) else t
                for t in batch)
            with torch.no_grad():
                #  inputs = self.package_inputs_from_batch(args,
                #                                          batch,
                #                                          known_labels=False)
                #  outputs = model(**inputs)
                #  self.on_predict_step(args, test_dataset, step, model, inputs,
                #                       outputs)
                self.on_predict_step(args, model, step, batch)

            pbar.update(step + 1)

        results = self.on_predict_end(args, test_features)
        return results



# -*- coding: utf-8 -*-
# File: meta.py


from .base import ImageAugmentor
from .transform import NoOpTransform, TransformList, TransformFactory

__all__ = ['RandomChooseAug', 'MapImage', 'Identity', 'RandomApplyAug',
           'RandomOrderAug']


class Identity(ImageAugmentor):
    """ A no-op augmentor """
    def get_transform(self, img):
        return NoOpTransform()


class RandomApplyAug(ImageAugmentor):
    """ Randomly apply the augmentor with a probability.
        Otherwise do nothing
    """

    def __init__(self, aug, prob):
        """
        Args:
            aug (ImageAugmentor): an augmentor.
            prob (float): the probability to apply the augmentor.
        """
        self._init(locals())
        super(RandomApplyAug, self).__init__()

    def get_transform(self, img):
        p = self.rng.rand()
        if p < self.prob:
            return self.aug.get_transform(img)
        else:
            return NoOpTransform()

    def reset_state(self):
        super(RandomApplyAug, self).reset_state()
        self.aug.reset_state()


class RandomChooseAug(ImageAugmentor):
    """ Randomly choose one from a list of augmentors """
    def __init__(self, aug_lists):
        """
        Args:
            aug_lists (list): list of augmentors, or list of (augmentor, probability) tuples
        """
        if isinstance(aug_lists[0], (tuple, list)):
            prob = [k[1] for k in aug_lists]
            aug_lists = [k[0] for k in aug_lists]
            self._init(locals())
        else:
            prob = [1.0 / len(aug_lists)] * len(aug_lists)
            self._init(locals())
        super(RandomChooseAug, self).__init__()

    def reset_state(self):
        super(RandomChooseAug, self).reset_state()
        for a in self.aug_lists:
            a.reset_state()

    def get_transform(self, img):
        aug_idx = self.rng.choice(len(self.aug_lists), p=self.prob)
        return self.aug_lists[aug_idx].get_transform(img)


class RandomOrderAug(ImageAugmentor):
    """
    Apply the augmentors with randomized order.
    """

    def __init__(self, aug_lists):
        """
        Args:
            aug_lists (list): list of augmentors.
                The augmentors are assumed to not change the shape of images.
        """
        self._init(locals())
        super(RandomOrderAug, self).__init__()

    def reset_state(self):
        super(RandomOrderAug, self).reset_state()
        for a in self.aug_lists:
            a.reset_state()

    def get_transform(self, img):
        # Note: this makes assumption that the augmentors do not make changes
        # to the image that will affect how the transforms will be instantiated
        # in the subsequent augmentors.
        idxs = self.rng.permutation(len(self.aug_lists))
        tfms = [self.aug_lists[k].get_transform(img)
                for k in range(len(self.aug_lists))]
        return TransformList([tfms[k] for k in idxs])


class MapImage(ImageAugmentor):
    """
    Map the image array by simple functions.
    """

    def __init__(self, func, coord_func=None):
        """
        Args:
            func: a function which takes an image array and return an augmented one
            coord_func: optional. A function which takes coordinates and return augmented ones.
                Coordinates should be Nx2 array of (x, y)s.
        """
        super(MapImage, self).__init__()
        self.func = func
        self.coord_func = coord_func

    def get_transform(self, img):
        if self.coord_func:
            return TransformFactory(name="MapImage", apply_image=self.func, apply_coords=self.coord_func)
        else:
            return TransformFactory(name="MapImage", apply_image=self.func)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import math
import torch
import torch.nn as nn
from torchcrf import CRF
from itertools import repeat
from transformers import BertModel
from .ner_decodes import crf_decode, span_decode


def vote(entities_list, threshold=0.9):
    """
    实体级别的投票方式  (entity_type, entity_start, entity_end, entity_text)
    :param entities_list: 所有模型预测出的一个文件的实体
    :param threshold:大于70%模型预测出来的实体才能被选中
    :return:
    """
    threshold_nums = int(len(entities_list) * threshold)
    entities_dict = defaultdict(int)
    entities = defaultdict(list)

    for _entities in entities_list:
        for _type in _entities:
            for _ent in _entities[_type]:
                entities_dict[(_type, _ent[0], _ent[1])] += 1

    for key in entities_dict:
        if entities_dict[key] >= threshold_nums:
            entities[key[0]].append((key[1], key[2]))

    return entities


def ensemble_vote(entities_list, threshold=0.9):
    """
    针对 ensemble model 进行的 vote
    实体级别的投票方式  (entity_type, entity_start, entity_end, entity_text)
    """
    threshold_nums = int(len(entities_list) * threshold)
    entities_dict = defaultdict(int)

    entities = defaultdict(list)

    for _entities in entities_list:
        for _id in _entities:
            for _ent in _entities[_id]:
                entities_dict[(_id, ) + _ent] += 1

    for key in entities_dict:
        if entities_dict[key] >= threshold_nums:
            entities[key[0]].append(key[1:])

    return entities


class BaseModel(nn.Module):
    def __init__(self, bert_dir, dropout_prob):
        super(BaseModel, self).__init__()
        config_path = os.path.join(bert_dir, 'config.json')

        assert os.path.exists(bert_dir) and os.path.exists(config_path), \
            'pretrained bert file does not exist'

        self.bert_module = BertModel.from_pretrained(
            bert_dir,
            output_hidden_states=True,
            hidden_dropout_prob=dropout_prob)

        self.bert_config = self.bert_module.config

    @staticmethod
    def _init_weights(blocks, **kwargs):
        """
        参数初始化，将 Linear / Embedding / LayerNorm 与 Bert 进行一样的初始化
        """
        for block in blocks:
            for module in block.modules():
                if isinstance(module, nn.Linear):
                    if module.bias is not None:
                        nn.init.zeros_(module.bias)
                elif isinstance(module, nn.Embedding):
                    nn.init.normal_(module.weight,
                                    mean=0,
                                    std=kwargs.pop('initializer_range', 0.02))
                elif isinstance(module, nn.LayerNorm):
                    nn.init.ones_(module.weight)
                    nn.init.zeros_(module.bias)


# baseline
class CRFModel(BaseModel):
    def __init__(self, bert_dir, num_tags, dropout_prob=0.1, **kwargs):
        super(CRFModel, self).__init__(bert_dir=bert_dir,
                                       dropout_prob=dropout_prob)

        out_dims = self.bert_config.hidden_size

        mid_linear_dims = kwargs.pop('mid_linear_dims', 128)

        self.mid_linear = nn.Sequential(nn.Linear(out_dims, mid_linear_dims),
                                        nn.ReLU(), nn.Dropout(dropout_prob))

        out_dims = mid_linear_dims

        self.classifier = nn.Linear(out_dims, num_tags)

        self.loss_weight = nn.Parameter(torch.FloatTensor(1),
                                        requires_grad=True)
        self.loss_weight.data.fill_(-0.2)

        self.crf_module = CRF(num_tags=num_tags, batch_first=True)

        init_blocks = [self.mid_linear, self.classifier]

        self._init_weights(
            init_blocks, initializer_range=self.bert_config.initializer_range)

    def forward(self,
                token_ids,
                attention_masks,
                token_type_ids,
                labels=None,
                pseudo=None):

        bert_outputs = self.bert_module(input_ids=token_ids,
                                        attention_mask=attention_masks,
                                        token_type_ids=token_type_ids)

        # 常规
        seq_out = bert_outputs[0]

        seq_out = self.mid_linear(seq_out)

        emissions = self.classifier(seq_out)

        if labels is not None:
            if pseudo is not None:
                # (batch,)
                tokens_loss = -1. * self.crf_module(
                    emissions=emissions,
                    tags=labels.long(),
                    mask=attention_masks.byte(),
                    reduction='none')

                # nums of pseudo data
                pseudo_nums = pseudo.sum().item()
                total_nums = token_ids.shape[0]

                # learning parameter
                rate = torch.sigmoid(self.loss_weight)
                if pseudo_nums == 0:
                    loss_0 = tokens_loss.mean()
                    loss_1 = (rate * pseudo * tokens_loss).sum()
                else:
                    if total_nums == pseudo_nums:
                        loss_0 = 0
                    else:
                        loss_0 = ((1 - rate) *
                                  (1 - pseudo) * tokens_loss).sum() / (
                                      total_nums - pseudo_nums)
                    loss_1 = (rate * pseudo * tokens_loss).sum() / pseudo_nums

                tokens_loss = loss_0 + loss_1

            else:
                tokens_loss = -1. * self.crf_module(
                    emissions=emissions,
                    tags=labels.long(),
                    mask=attention_masks.byte(),
                    reduction='mean')

            out = (tokens_loss, )

        else:
            tokens_out = self.crf_module.decode(emissions=emissions,
                                                mask=attention_masks.byte())

            out = (tokens_out, emissions)

        return out


class SpanModel(BaseModel):
    def __init__(self,
                 bert_dir,
                 num_tags,
                 dropout_prob=0.1,
                 loss_type='ce',
                 **kwargs):
        """
        tag the subject and object corresponding to the predicate
        :param loss_type: train loss type in ['ce', 'ls_ce', 'focal']
        """
        super(SpanModel, self).__init__(bert_dir, dropout_prob=dropout_prob)

        out_dims = self.bert_config.hidden_size

        mid_linear_dims = kwargs.pop('mid_linear_dims', 128)

        self.num_tags = num_tags

        self.mid_linear = nn.Sequential(nn.Linear(out_dims, mid_linear_dims),
                                        nn.ReLU(), nn.Dropout(dropout_prob))

        out_dims = mid_linear_dims

        self.start_fc = nn.Linear(out_dims, num_tags)
        self.end_fc = nn.Linear(out_dims, num_tags)

        reduction = 'none'
        if loss_type == 'ce':
            self.criterion = nn.CrossEntropyLoss(reduction=reduction)
        elif loss_type == 'ls_ce':
            self.criterion = LabelSmoothingCrossEntropy(reduction=reduction)
        else:
            self.criterion = FocalLoss(reduction=reduction)

        self.loss_weight = nn.Parameter(torch.FloatTensor(1),
                                        requires_grad=True)
        self.loss_weight.data.fill_(-0.2)

        init_blocks = [self.mid_linear, self.start_fc, self.end_fc]

        self._init_weights(init_blocks)

    def forward(self,
                token_ids,
                attention_masks,
                token_type_ids,
                start_ids=None,
                end_ids=None,
                pseudo=None):

        bert_outputs = self.bert_module(input_ids=token_ids,
                                        attention_mask=attention_masks,
                                        token_type_ids=token_type_ids)

        seq_out = bert_outputs[0]

        seq_out = self.mid_linear(seq_out)

        start_logits = self.start_fc(seq_out)
        end_logits = self.end_fc(seq_out)

        out = (
            start_logits,
            end_logits,
        )

        if start_ids is not None and end_ids is not None and self.training:

            start_logits = start_logits.view(-1, self.num_tags)
            end_logits = end_logits.view(-1, self.num_tags)

            # 去掉 padding 部分的标签，计算真实 loss
            active_loss = attention_masks.view(-1) == 1
            active_start_logits = start_logits[active_loss]
            active_end_logits = end_logits[active_loss]

            active_start_labels = start_ids.view(-1)[active_loss]
            active_end_labels = end_ids.view(-1)[active_loss]

            if pseudo is not None:
                # (batch,)
                start_loss = self.criterion(
                    start_logits, start_ids.view(-1)).view(-1,
                                                           512).mean(dim=-1)
                end_loss = self.criterion(end_logits, end_ids.view(-1)).view(
                    -1, 512).mean(dim=-1)

                # nums of pseudo data
                pseudo_nums = pseudo.sum().item()
                total_nums = token_ids.shape[0]

                # learning parameter
                rate = torch.sigmoid(self.loss_weight)
                if pseudo_nums == 0:
                    start_loss = start_loss.mean()
                    end_loss = end_loss.mean()
                else:
                    if total_nums == pseudo_nums:
                        start_loss = (rate * pseudo *
                                      start_loss).sum() / pseudo_nums
                        end_loss = (rate * pseudo *
                                    end_loss).sum() / pseudo_nums
                    else:
                        start_loss = (rate*pseudo*start_loss).sum() / pseudo_nums \
                                     + ((1 - rate) * (1 - pseudo) * start_loss).sum() / (total_nums - pseudo_nums)
                        end_loss = (rate*pseudo*end_loss).sum() / pseudo_nums \
                                     + ((1 - rate) * (1 - pseudo) * end_loss).sum() / (total_nums - pseudo_nums)
            else:
                start_loss = self.criterion(active_start_logits,
                                            active_start_labels)
                end_loss = self.criterion(active_end_logits, active_end_labels)

            loss = start_loss + end_loss

            out = (loss, ) + out

        return out


class MRCModel(BaseModel):
    def __init__(self,
                 bert_dir,
                 dropout_prob=0.1,
                 use_type_embed=False,
                 loss_type='ce',
                 **kwargs):
        """
        tag the subject and object corresponding to the predicate
        :param use_type_embed: type embedding for the sentence
        :param loss_type: train loss type in ['ce', 'ls_ce', 'focal']
        """
        super(MRCModel, self).__init__(bert_dir, dropout_prob=dropout_prob)

        self.use_type_embed = use_type_embed
        self.use_smooth = loss_type

        out_dims = self.bert_config.hidden_size

        if self.use_type_embed:
            embed_dims = kwargs.pop('predicate_embed_dims',
                                    self.bert_config.hidden_size)
            self.type_embedding = nn.Embedding(13, embed_dims)

            self.conditional_layer_norm = ConditionalLayerNorm(
                out_dims, embed_dims, eps=self.bert_config.layer_norm_eps)

        mid_linear_dims = kwargs.pop('mid_linear_dims', 128)

        self.mid_linear = nn.Sequential(nn.Linear(out_dims, mid_linear_dims),
                                        nn.ReLU(), nn.Dropout(dropout_prob))

        out_dims = mid_linear_dims

        self.start_fc = nn.Linear(out_dims, 2)
        self.end_fc = nn.Linear(out_dims, 2)

        reduction = 'none'
        if loss_type == 'ce':
            self.criterion = nn.CrossEntropyLoss(reduction=reduction)
        elif loss_type == 'ls_ce':
            self.criterion = LabelSmoothingCrossEntropy(reduction=reduction)
        else:
            self.criterion = FocalLoss(reduction=reduction)

        self.loss_weight = nn.Parameter(torch.FloatTensor(1),
                                        requires_grad=True)
        self.loss_weight.data.fill_(-0.2)

        init_blocks = [self.mid_linear, self.start_fc, self.end_fc]

        if self.use_type_embed:
            init_blocks.append(self.type_embedding)

        self._init_weights(init_blocks)

    def forward(self,
                token_ids,
                attention_masks,
                token_type_ids,
                ent_type=None,
                start_ids=None,
                end_ids=None,
                pseudo=None):

        bert_outputs = self.bert_module(input_ids=token_ids,
                                        attention_mask=attention_masks,
                                        token_type_ids=token_type_ids)

        seq_out = bert_outputs[0]

        if self.use_type_embed:
            assert ent_type is not None, \
                'Using predicate embedding, predicate should be implemented'

            predicate_feature = self.type_embedding(ent_type)
            seq_out = self.conditional_layer_norm(seq_out, predicate_feature)

        seq_out = self.mid_linear(seq_out)

        start_logits = self.start_fc(seq_out)
        end_logits = self.end_fc(seq_out)

        out = (
            start_logits,
            end_logits,
        )

        if start_ids is not None and end_ids is not None:
            start_logits = start_logits.view(-1, 2)
            end_logits = end_logits.view(-1, 2)

            # 去掉 text_a 和 padding 部分的标签，计算真实 loss
            active_loss = token_type_ids.view(-1) == 1
            active_start_logits = start_logits[active_loss]
            active_end_logits = end_logits[active_loss]

            active_start_labels = start_ids.view(-1)[active_loss]
            active_end_labels = end_ids.view(-1)[active_loss]

            if pseudo is not None:
                # (batch,)
                start_loss = self.criterion(
                    start_logits, start_ids.view(-1)).view(-1,
                                                           512).mean(dim=-1)
                end_loss = self.criterion(end_logits, end_ids.view(-1)).view(
                    -1, 512).mean(dim=-1)

                # nums of pseudo data
                pseudo_nums = pseudo.sum().item()
                total_nums = token_ids.shape[0]

                # learning parameter
                rate = torch.sigmoid(self.loss_weight)
                if pseudo_nums == 0:
                    start_loss = start_loss.mean()
                    end_loss = end_loss.mean()
                else:
                    if total_nums == pseudo_nums:
                        start_loss = (rate * pseudo *
                                      start_loss).sum() / pseudo_nums
                        end_loss = (rate * pseudo *
                                    end_loss).sum() / pseudo_nums
                    else:
                        start_loss = (rate*pseudo*start_loss).sum() / pseudo_nums \
                                     + ((1 - rate) * (1 - pseudo) * start_loss).sum() / (total_nums - pseudo_nums)
                        end_loss = (rate*pseudo*end_loss).sum() / pseudo_nums \
                                     + ((1 - rate) * (1 - pseudo) * end_loss).sum() / (total_nums - pseudo_nums)
            else:
                start_loss = self.criterion(active_start_logits,
                                            active_start_labels)
                end_loss = self.criterion(active_end_logits, active_end_labels)

            loss = start_loss + end_loss

            out = (loss, ) + out

        return out


class EnsembleCRFModel:
    def __init__(self,
                 model_path_list,
                 bert_dir_list,
                 num_tags,
                 device,
                 lamb=1 / 3):

        self.models = []
        self.crf_module = CRF(num_tags=num_tags, batch_first=True)
        self.lamb = lamb

        for idx, _path in enumerate(model_path_list):
            print(f'Load model from {_path}')

            print(f'Load model type: {bert_dir_list[0]}')
            model = CRFModel(bert_dir=bert_dir_list[0], num_tags=num_tags)

            model.load_state_dict(
                torch.load(_path, map_location=torch.device('cpu')))

            model.eval()
            model.to(device)

            self.models.append(model)
            if idx == 0:
                print(f'Load CRF weight from {_path}')
                self.crf_module.load_state_dict(model.crf_module.state_dict())
                self.crf_module.to(device)

    def weight(self, t):
        """
        牛顿冷却定律加权融合
        """
        return math.exp(-self.lamb * t)

    def predict(self, model_inputs):
        weight_sum = 0.
        logits = None
        attention_masks = model_inputs['attention_masks']

        for idx, model in enumerate(self.models):
            # 使用牛顿冷却概率融合
            weight = self.weight(idx)

            # 使用概率平均融合
            # weight = 1 / len(self.models)

            tmp_logits = model(**model_inputs)[1] * weight
            weight_sum += weight

            if logits is None:
                logits = tmp_logits
            else:
                logits += tmp_logits

        logits = logits / weight_sum

        tokens_out = self.crf_module.decode(emissions=logits,
                                            mask=attention_masks.byte())

        return tokens_out

    def vote_entities(self, model_inputs, sent, id2ent, threshold):
        entities_ls = []
        for idx, model in enumerate(self.models):
            tmp_tokens = model(**model_inputs)[0][0]
            tmp_entities = crf_decode(tmp_tokens, sent, id2ent)
            entities_ls.append(tmp_entities)

        return vote(entities_ls, threshold)


class EnsembleSpanModel:
    def __init__(self, model_path_list, bert_dir_list, num_tags, device):

        self.models = []

        for idx, _path in enumerate(model_path_list):
            print(f'Load model from {_path}')

            print(f'Load model type: {bert_dir_list[0]}')
            model = SpanModel(bert_dir=bert_dir_list[0], num_tags=num_tags)

            model.load_state_dict(
                torch.load(_path, map_location=torch.device('cpu')))

            model.eval()
            model.to(device)

            self.models.append(model)

    def predict(self, model_inputs):
        start_logits, end_logits = None, None

        for idx, model in enumerate(self.models):

            # 使用概率平均融合
            weight = 1 / len(self.models)

            tmp_start_logits, tmp_end_logits = model(**model_inputs)

            tmp_start_logits = tmp_start_logits * weight
            tmp_end_logits = tmp_end_logits * weight

            if start_logits is None:
                start_logits = tmp_start_logits
                end_logits = tmp_end_logits
            else:
                start_logits += tmp_start_logits
                end_logits += tmp_end_logits

        return start_logits, end_logits

    def vote_entities(self, model_inputs, sent, id2ent, threshold):
        entities_ls = []

        for idx, model in enumerate(self.models):

            start_logits, end_logits = model(**model_inputs)
            start_logits = start_logits[0].cpu().numpy()[1:1 + len(sent)]
            end_logits = end_logits[0].cpu().numpy()[1:1 + len(sent)]

            decode_entities = span_decode(start_logits, end_logits, sent,
                                          id2ent)

            entities_ls.append(decode_entities)

        return vote(entities_ls, threshold)



# -*- coding: utf-8 -*-
# File: develop.py
# Author: tensorpack contributors


""" Utilities for developers only.
These are not visible to users (not automatically imported). And should not
appeared in docs."""
import functools
import importlib
import os
import types
from collections import defaultdict
from datetime import datetime
import six

from . import logger

__all__ = []


def create_dummy_class(klass, dependency):
    """
    When a dependency of a class is not available, create a dummy class which throws ImportError when used.

    Args:
        klass (str): name of the class.
        dependency (str): name of the dependency.

    Returns:
        class: a class object
    """
    assert not building_rtfd()

    class _DummyMetaClass(type):
        # throw error on class attribute access
        def __getattr__(_, __):
            raise AttributeError("Cannot import '{}', therefore '{}' is not available".format(dependency, klass))

    @six.add_metaclass(_DummyMetaClass)
    class _Dummy(object):
        # throw error on constructor
        def __init__(self, *args, **kwargs):
            raise ImportError("Cannot import '{}', therefore '{}' is not available".format(dependency, klass))

    return _Dummy


def create_dummy_func(func, dependency):
    """
    When a dependency of a function is not available, create a dummy function which throws ImportError when used.

    Args:
        func (str): name of the function.
        dependency (str or list[str]): name(s) of the dependency.

    Returns:
        function: a function object
    """
    assert not building_rtfd()

    if isinstance(dependency, (list, tuple)):
        dependency = ','.join(dependency)

    def _dummy(*args, **kwargs):
        raise ImportError("Cannot import '{}', therefore '{}' is not available".format(dependency, func))
    return _dummy


def building_rtfd():
    """
    Returns:
        bool: if the library is being imported to generate docs now.
    """
    return os.environ.get('READTHEDOCS') == 'True' \
        or os.environ.get('DOC_BUILDING')


_DEPRECATED_LOG_NUM = defaultdict(int)


def log_deprecated(name="", text="", eos="", max_num_warnings=None):
    """
    Log deprecation warning.

    Args:
        name (str): name of the deprecated item.
        text (str, optional): information about the deprecation.
        eos (str, optional): end of service date such as "YYYY-MM-DD".
        max_num_warnings (int, optional): the maximum number of times to print this warning
    """
    assert name or text
    if eos:
        eos = "after " + datetime(*map(int, eos.split("-"))).strftime("%d %b")
    if name:
        if eos:
            warn_msg = "%s will be deprecated %s. %s" % (name, eos, text)
        else:
            warn_msg = "%s was deprecated. %s" % (name, text)
    else:
        warn_msg = text
        if eos:
            warn_msg += " Legacy period ends %s" % eos

    if max_num_warnings is not None:
        if _DEPRECATED_LOG_NUM[warn_msg] >= max_num_warnings:
            return
        _DEPRECATED_LOG_NUM[warn_msg] += 1
    logger.warn("[Deprecated] " + warn_msg)


def deprecated(text="", eos="", max_num_warnings=None):
    """
    Args:
        text, eos, max_num_warnings: same as :func:`log_deprecated`.

    Returns:
        a decorator which deprecates the function.

    Example:
        .. code-block:: python

            @deprecated("Explanation of what to do instead.", "2017-11-4")
            def foo(...):
                pass
    """

    def get_location():
        import inspect
        frame = inspect.currentframe()
        if frame:
            callstack = inspect.getouterframes(frame)[-1]
            return '%s:%i' % (callstack[1], callstack[2])
        else:
            stack = inspect.stack(0)
            entry = stack[2]
            return '%s:%i' % (entry[1], entry[2])

    def deprecated_inner(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            name = "{} [{}]".format(func.__name__, get_location())
            log_deprecated(name, text, eos, max_num_warnings=max_num_warnings)
            return func(*args, **kwargs)
        return new_func
    return deprecated_inner


def HIDE_DOC(func):
    func.__HIDE_SPHINX_DOC__ = True
    return func


# Copied from https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/util/lazy_loader.py
class LazyLoader(types.ModuleType):
    def __init__(self, local_name, parent_module_globals, name):
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals
        super(LazyLoader, self).__init__(name)

    def _load(self):
        # Import the target module and insert it into the parent's namespace
        module = importlib.import_module(self.__name__)
        self._parent_module_globals[self._local_name] = module

        # Update this object's dict so that if someone keeps a reference to the
        #   LazyLoader, lookups are efficient (__getattr__ is only called on lookups
        #   that fail).
        self.__dict__.update(module.__dict__)

        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)
